# Copyright 2019 The Texar Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Base class for decoders.
"""

import copy
from abc import ABC, abstractmethod
from typing import Callable, Generic, Optional, Tuple, TypeVar, Union, overload

import torch
from torch import nn

from texar.torch.core.layers import identity, Identity
from texar.torch.module_base import ModuleBase
from texar.torch.modules.decoders import decoder_helpers as helpers
from texar.torch.modules.decoders.decoder_helpers import Helper
from texar.torch.utils import utils
from texar.torch.utils.dtypes import torch_bool

__all__ = [
    '_make_output_layer',
    'DecoderBase',
]

State = TypeVar('State')
Output = TypeVar('Output')  # output type can be of any nested structure


def _make_output_layer(layer: Optional[Union[nn.Module, torch.Tensor]],
                       vocab_size: Optional[int],
                       output_size: int,
                       bias: bool) -> Tuple[nn.Module, Optional[int]]:
    r"""Construct the output layer for decoders. Based on the input, multiple
    types of output layers could be constructed:

    - If ``layer`` is a :torch_nn:`Module`, then the layer is returned as is.
    - If ``layer`` is `None`, then a :torch_nn:`Linear` layer is constructed
      with ``output_size`` and ``vocab_size`` as input and output dimensions.
    - If ``layer`` is a :tensor:`Tensor`, then a :torch_nn:`Linear` layer is
      constructed with the provided tensor as parameters. Note that this tensor
      should have transposed shape, i.e. shape of ``[vocab_size, output_size]``.
      Also, if the provided tensor is not an instance of :torch_nn:`Parameter`,
      it will **not** accumulate gradients.
    - If ``layer`` is :method:`texar.torch.core.identity`, identity function is
      used as the output layer.
    """
    if isinstance(layer, nn.Module):
        output_layer = layer
    elif layer is None:
        if vocab_size is None:
            raise ValueError(
                "Either `output_layer` or `vocab_size` must be provided. "
                "Set `output_layer=tx.core.identity` if no output "
                "layer is wanted.")
        output_layer = nn.Linear(output_size, vocab_size, bias)
    elif torch.is_tensor(layer):
        vocab_size = layer.size(0)
        output_layer = nn.Linear(layer.size(1), vocab_size, bias)
        if not isinstance(layer, nn.Parameter):
            layer = nn.Parameter(layer, requires_grad=False)
        output_layer.weight = layer
    elif layer is identity:
        output_layer = Identity()
    else:
        raise ValueError(
            f"output_layer should be an instance of `nn.Module`, a tensor,"
            f"or None. Unsupported type: {type(layer)}")

    return output_layer, vocab_size


TokenEmbedder = Union[nn.Module, Callable[[torch.LongTensor], torch.Tensor]]
TokenPosEmbedder = Union[
    nn.Module, Callable[[torch.LongTensor, torch.LongTensor], torch.Tensor]]


class DecoderBase(ModuleBase, Generic[State, Output], ABC):
    r"""Base class inherited by all RNN decoder classes.
    See :class:`~texar.torch.modules.BasicRNNDecoder` for the arguments.

    See :meth:`forward` for the inputs and outputs of RNN decoders in general.
    """

    def __init__(self,
                 token_embedder: Optional[TokenEmbedder] = None,
                 token_pos_embedder: Optional[TokenPosEmbedder] = None,
                 input_time_major: bool = False,
                 output_time_major: bool = False,
                 hparams=None):
        super().__init__(hparams=hparams)

        self._train_helper: Optional[Helper] = None
        self._infer_helper: Optional[Helper] = None
        self._input_time_major = input_time_major
        self._output_time_major = output_time_major

        if (token_embedder is not None and
                token_pos_embedder is not None):
            raise ValueError("At most one among `token_embedder` and "
                             "`token_pos_embedder` should be specified")
        if token_embedder is None and token_pos_embedder is None:
            embed_token_func = self.embed_tokens.__func__  # type: ignore
            if embed_token_func is DecoderBase.embed_tokens:
                raise ValueError(
                    "Either `token_embedder` or `token_pos_embedder` must not "
                    "be `None` if `DecoderBase.embed_tokens` is not "
                    "overridden.")

        self._token_embedder = token_embedder
        self._token_pos_embedder = token_pos_embedder

    def embed_tokens(self, tokens: torch.LongTensor,
                     positions: torch.LongTensor) -> torch.Tensor:
        r"""Convert tokens along with positions to embeddings.

        Args:
            tokens: A :tensor:`LongTensor` denoting the token indices to convert
                to embeddings.
            positions: A :tensor:`LongTensor` with the same size as
                :attr:`tokens`, denoting the positions of the tokens. This is
                useful if the decoder uses positional embeddings.

        Returns:
            A :tensor:`Tensor` of size ``tokens.size() + (embed_dim,)``,
            denoting the converted embeddings.
        """
        if self._token_embedder is not None:
            return self._token_embedder(tokens)
        assert self._token_pos_embedder is not None
        return self._token_pos_embedder(tokens, positions)

    def create_helper(self, *,
                      decoding_strategy: Optional[str] = None,
                      start_tokens: Optional[torch.LongTensor] = None,
                      end_token: Optional[int] = None,
                      softmax_temperature: Optional[float] = None,
                      infer_mode: Optional[bool] = None,
                      **kwargs) -> Helper:
        r"""Create a helper instance for the decoder. This is a shared interface
        for both :class:`~texar.torch.modules.BasicRNNDecoder` and
        :class:`~texar.torch.modules.AttentionRNNDecoder`.

        The function provides **3 ways** to specify the
        decoding method, with varying flexibility:

        1. The :attr:`decoding_strategy` argument: A string taking value of:

            - **"train_greedy"**: decoding in teacher-forcing fashion (i.e.,
              feeding `ground truth` to decode the next step), and each sample
              is obtained by taking the `argmax` of the output logits.
              Arguments :attr:`(inputs, sequence_length)`
              are required for this strategy, and argument :attr:`embedding`
              is optional.
            - **"infer_greedy"**: decoding in inference fashion (i.e., feeding
              the `generated` sample to decode the next step), and each sample
              is obtained by taking the `argmax` of the output logits.
              Arguments :attr:`(embedding, start_tokens, end_token)` are
              required for this strategy, and argument
              :attr:`max_decoding_length` is optional.
            - **"infer_sample"**: decoding in inference fashion, and each
              sample is obtained by `random sampling` from the RNN output
              distribution. Arguments
              :attr:`(embedding, start_tokens, end_token)` are
              required for this strategy, and argument
              :attr:`max_decoding_length` is optional.

          This argument is used only when argument :attr:`helper` is `None`.

          Example:

            .. code-block:: python

                embedder = WordEmbedder(vocab_size=data.vocab.size)
                decoder = BasicRNNDecoder(vocab_size=data.vocab.size)

                # Teacher-forcing decoding
                outputs_1, _, _ = decoder(
                    decoding_strategy='train_greedy',
                    inputs=embedder(data_batch['text_ids']),
                    sequence_length=data_batch['length'] - 1)

                # Random sample decoding. Gets 100 sequence samples
                outputs_2, _, sequence_length = decoder(
                    decoding_strategy='infer_sample',
                    start_tokens=[data.vocab.bos_token_id] * 100,
                    end_token=data.vocab.eos.token_id,
                    embedding=embedder,
                    max_decoding_length=60)

        2. The :attr:`helper` argument: An instance of subclass of
           :class:`~texar.torch.modules.decoders.decoder_helpers.Helper`. This
           provides a superset of decoding strategies than above, for example:

            - :class:`~texar.torch.modules.TrainingHelper` corresponding to the
              "train_greedy" strategy.
            - :class:`~texar.torch.modules.ScheduledEmbeddingTrainingHelper` and
              :class:`~texar.torch.modules.ScheduledOutputTrainingHelper` for
              scheduled sampling.
            - :class:`~texar.torch.modules.SoftmaxEmbeddingHelper` and
              :class:`~texar.torch.modules.GumbelSoftmaxEmbeddingHelper` for
              soft decoding and gradient backpropagation.

          This means gives the maximal flexibility of configuring the decoding
          strategy.

          Example:

            .. code-block:: python

                embedder = WordEmbedder(vocab_size=data.vocab.size)
                decoder = BasicRNNDecoder(vocab_size=data.vocab.size)

                # Teacher-forcing decoding, same as above with
                # `decoding_strategy='train_greedy'`
                helper_1 = TrainingHelper(
                    inputs=embedders(data_batch['text_ids']),
                    sequence_length=data_batch['length'] - 1)
                outputs_1, _, _ = decoder(helper=helper_1)

                # Gumbel-softmax decoding
                helper_2 = GumbelSoftmaxEmbeddingHelper(
                    embedding=embedder,
                    start_tokens=[data.vocab.bos_token_id] * 100,
                    end_token=data.vocab.eos_token_id,
                    tau=0.1)
                outputs_2, _, sequence_length = decoder(
                    max_decoding_length=60, helper=helper_2)

        3. ``hparams["helper_train"]`` and ``hparams["helper_infer"]``:
           Specifying the helper through hyperparameters. Train and infer
           strategy is toggled based on :attr:`mode`. Appropriate arguments
           (e.g., :attr:`inputs`, :attr:`start_tokens`, etc) are selected to
           construct the helper. Additional arguments for helper constructor
           can be provided either through :attr:`**kwargs`, or through
           ``hparams["helper_train/infer"]["kwargs"]``.

           This means is used only when both :attr:`decoding_strategy` and
           :attr:`helper` are `None`.

           Example:

             .. code-block:: python

                h = {
                    "helper_infer": {
                        "type": "GumbelSoftmaxEmbeddingHelper",
                        "kwargs": { "tau": 0.1 }
                    }
                }
                embedder = WordEmbedder(vocab_size=data.vocab.size)
                decoder = BasicRNNDecoder(vocab_size=data.vocab.size, hparams=h)

                # Gumbel-softmax decoding
                decoder.eval()  # disable dropout
                output, _, _ = decoder(
                    decoding_strategy=None, # Sets to None explicit
                    embedding=embedder,
                    start_tokens=[data.vocab.bos_token_id] * 100,
                    end_token=data.vocab.eos_token_id,
                    max_decoding_length=60)

        Args:
            decoding_strategy (str): A string specifying the decoding
                strategy. Different arguments are required based on the
                strategy.
                Ignored if :attr:`helper` is given.
            start_tokens (optional): A :tensor:`LongTensor` of shape
                ``[batch_size]``, the start tokens.
                Used when :attr:`decoding_strategy` is ``"infer_greedy"`` or
                ``"infer_sample"``, or when `hparams`-configured
                helper is used.
                When used with the Texar data module, to get ``batch_size``
                samples where ``batch_size`` is changing according to the data
                module, this can be set as
                ``start_tokens=torch.full_like(batch['length'], bos_token_id)``.
            end_token (optional): A integer or 0D :tensor:`LongTensor`, the
                token that marks the end of decoding.
                Used when :attr:`decoding_strategy` is ``"infer_greedy"`` or
                ``"infer_sample"``, or when `hparams`-configured helper is used.
            softmax_temperature (float, optional): Value to divide the logits
                by before computing the softmax. Larger values (above 1.0)
                result in more random samples. Must be > 0. If `None`, 1.0 is
                used. Used when ``decoding_strategy="infer_sample"``.
            infer_mode (optional): If not `None`, overrides mode given by
                :attr:`self.training`.
            **kwargs: Other keyword arguments for constructing helpers
                defined by ``hparams["helper_train"]`` or
                ``hparams["helper_infer"]``.

        Returns:
            The constructed helper instance.
        """
        if decoding_strategy is not None:
            if decoding_strategy == 'train_greedy':
                helper: Helper = helpers.TrainingHelper(
                    self._input_time_major)
            elif decoding_strategy in ['infer_greedy', 'infer_sample']:
                if start_tokens is None or end_token is None:
                    raise ValueError(
                        f"When using '{decoding_strategy}' decoding strategy, "
                        f"'embedding', 'start_tokens', and 'end_token' must "
                        f"not be `None`.")
                if decoding_strategy == 'infer_greedy':
                    helper = helpers.GreedyEmbeddingHelper(
                        start_tokens, end_token)
                else:
                    helper = helpers.SampleEmbeddingHelper(
                        start_tokens, end_token, softmax_temperature)
            else:
                raise ValueError(
                    f"Unknown decoding strategy: {decoding_strategy}")
        else:
            is_training = (not infer_mode if infer_mode is not None
                           else self.training)
            if is_training:
                kwargs_ = copy.copy(self._hparams.helper_train.kwargs.todict())
                helper_type = self._hparams.helper_train.type
            else:
                kwargs_ = copy.copy(self._hparams.helper_infer.kwargs.todict())
                helper_type = self._hparams.helper_infer.type
            kwargs_.update({
                'time_major': self._input_time_major,
                'start_tokens': start_tokens,
                'end_token': end_token,
                'softmax_temperature': softmax_temperature})
            kwargs_.update(kwargs)
            helper = helpers.get_helper(helper_type, **kwargs_)
        return helper

    def _create_or_get_helper(self, infer_mode: Optional[bool] = None,
                              **kwargs) -> Helper:
        # Prefer creating a new helper when at least one kwarg is specified.
        prefer_new = (len(kwargs) > 0)
        kwargs.update(infer_mode=infer_mode)
        is_training = (not infer_mode if infer_mode is not None
                       else self.training)
        helper = self._train_helper if is_training else self._infer_helper
        if prefer_new or helper is None:
            helper = self.create_helper(**kwargs)
            if is_training and self._train_helper is None:
                self._train_helper = helper
            elif not is_training and self._infer_helper is None:
                self._infer_helper = helper
        return helper

    def set_default_train_helper(self, helper: Helper):
        r"""Set the default helper used in training mode.

        Args:
            helper: The helper to set as default training helper.
        """
        self._train_helper = helper

    def set_default_infer_helper(self, helper: Helper):
        r"""Set the default helper used in eval (inference) mode.

        Args:
            helper: The helper to set as default inference helper.
        """
        self._infer_helper = helper

    def dynamic_decode(self, helper: Helper, inputs: Optional[torch.Tensor],
                       sequence_length: Optional[torch.LongTensor],
                       initial_state: Optional[State],
                       max_decoding_length: Optional[int] = None,
                       impute_finished: bool = False,
                       step_hook: Optional[Callable[[int], None]] = None) \
            -> Tuple[Output, Optional[State], torch.LongTensor]:
        r"""Generic routine for dynamic decoding. Please check the
        `documentation
        <https://www.tensorflow.org/api_docs/python/tf/contrib/seq2seq/dynamic_decode>`_
        for the TensorFlow counterpart.

        Returns:
            A tuple of output, final state, and sequence lengths. Note that
            final state could be `None`, when all sequences are of zero length
            and :attr:`initial_state` is also `None`.
        """

        # Decode
        finished, step_inputs, state = self.initialize(
            helper, inputs, sequence_length, initial_state)

        zero_outputs = step_inputs.new_zeros(
            step_inputs.size(0), self.output_size)

        if max_decoding_length is not None:
            finished |= (max_decoding_length <= 0)
        sequence_lengths = torch.zeros_like(
            finished, dtype=torch.long, device=finished.device)
        time = 0

        outputs = []

        while (not torch.all(finished).item() and
               (max_decoding_length is None or time < max_decoding_length)):

            next_outputs, decoder_state = \
                self.step(helper, time, step_inputs, state)

            if max_decoding_length is not None and \
                    time + 1 == max_decoding_length:
                # Maximum decoding length reached, mark all batches as finished.
                # This requires special handling because performing lookup on
                # position embeddings with `time + 1` may result in IndexError.
                decoder_finished = torch.tensor(1, dtype=torch_bool,
                                                device=finished.device)
                # Since `next_inputs` will not be used, simply create a null
                # tensor.
                next_inputs = torch.empty(0)
            else:
                next_inputs, decoder_finished = self.next_inputs(
                    helper, time, next_outputs)

            if getattr(self, 'tracks_own_finished', False):
                next_finished = decoder_finished
            else:
                next_finished = decoder_finished | finished

            # Zero out output values past finish
            if impute_finished:
                emit = utils.map_structure_zip(
                    lambda new, cur: torch.where(finished, cur, new),
                    (next_outputs, zero_outputs))
                next_state = utils.map_structure_zip(
                    lambda new, cur: torch.where(finished, cur, new),
                    (decoder_state, state))
            else:
                emit = next_outputs
                next_state = decoder_state

            outputs.append(emit)
            sequence_lengths.index_fill_(
                dim=0, value=time + 1,
                index=torch.nonzero((~finished).long()).flatten())
            time += 1
            finished = next_finished
            step_inputs = next_inputs
            state = next_state

            if step_hook is not None:
                step_hook(time)

        final_outputs = utils.map_structure_zip(
            lambda *tensors: torch.stack(tensors),
            outputs)  # output at each time step may be a namedtuple
        final_state = state
        final_sequence_lengths = sequence_lengths

        try:
            final_outputs, final_state = self.finalize(
                final_outputs, final_state, final_sequence_lengths)
        except NotImplementedError:
            pass

        if not self._output_time_major:
            final_outputs = utils.map_structure(
                lambda x: x.transpose(0, 1) if x.dim() >= 2 else x,
                final_outputs)

        return final_outputs, final_state, final_sequence_lengths

    @abstractmethod
    def initialize(self, helper: Helper, inputs: Optional[torch.Tensor],
                   sequence_length: Optional[torch.LongTensor],
                   initial_state: Optional[State]) \
            -> Tuple[torch.ByteTensor, torch.Tensor, Optional[State]]:
        r"""Called before any decoding iterations.

        This methods must compute initial input values and initial state.

        Args:
            helper: The :class:`~texar.torch.modules.Helper` instance to use.
            inputs (optional): A (structure of) input tensors.
            sequence_length (optional): A :tensor:`LongTensor` representing
                lengths of each sequence.
            initial_state: A possibly nested structure of tensors indicating the
                initial decoder state.

        Returns:
            A tuple ``(finished, initial_inputs, initial_state)`` representing
            initial values of ``finished`` flags, inputs, and state.
        """
        raise NotImplementedError

    @abstractmethod
    def step(self, helper: Helper, time: int,
             inputs: torch.Tensor, state: Optional[State]) \
            -> Tuple[Output, State]:
        r"""Compute the output and the state at the current time step.
        Called per step of decoding (but only once for dynamic decoding).

        Args:
            helper: The :class:`~texar.torch.modules.Helper` instance to use.
            time (int): Current step number.
            inputs: Inputs for this time step.
            state: Decoder state from the previous time step.

        Returns:
            A tuple ``(outputs, next_state)``.

            - ``outputs`` is an object containing the decoder output.
            - ``next_state`` is the decoder state for the next time step.
        """
        raise NotImplementedError

    @abstractmethod
    def next_inputs(self, helper: Helper, time: int, outputs: Output) -> \
            Tuple[torch.Tensor, torch.ByteTensor]:
        r"""Compute the input for the next time step.
        Called per step of decoding (but only once for dynamic decoding).

        Args:
            helper: The :class:`~texar.torch.modules.Helper` instance to use.
            time (int): Current step number.
            outputs: An object containing the decoder output.

        Returns:
            A tuple ``(next_inputs, finished)``.

            - ``next_inputs`` is the tensor that should be used as input for the
              next step.
            - ``finished`` is a :torch:`ByteTensor` tensor telling whether the
              sequence is complete, for each sequence in the batch.
        """
        raise NotImplementedError

    # TODO: Remove these once pylint supports function stubs.
    # pylint: disable=missing-docstring,unused-argument,no-self-use
    # pylint: disable=function-redefined

    @overload
    def finalize(self, outputs: Output, final_state: State,
                 sequence_lengths: torch.LongTensor) -> Tuple[Output, State]:
        ...

    @overload
    def finalize(self, outputs: Output, final_state: Optional[State],
                 sequence_lengths: torch.LongTensor) \
            -> Tuple[Output, Optional[State]]:
        ...

    def finalize(self, outputs, final_state, sequence_lengths):
        r"""Called after all decoding iterations have finished.

        Args:
            outputs: Outputs at each time step.
            final_state: The RNNCell state after the last time step.
            sequence_lengths: Sequence lengths for each sequence in batch.

        Returns:
            A tuple ``(outputs, final_state)``.

            - ``outputs`` is an object containing the decoder output.
            - ``final_state`` is the final decoder state.
        """
        return outputs, final_state

    # pylint: enable=missing-docstring,unused-argument,no-self-use
    # pylint: enable=function-redefined

    @property
    def vocab_size(self):
        r"""The vocabulary size.
        """
        return self._vocab_size

    @property
    def output_layer(self):
        r"""The output layer.
        """
        return self._output_layer



# Copyright 2019 The Texar Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Unit tests for T5 encoder-decoder
"""

import unittest

import numpy
import torch

from texar.torch.modules.encoder_decoders import T5EncoderDecoder
from texar.torch.utils.test import pretrained_test


class T5EncoderDecoderTest(unittest.TestCase):
    r"""Tests :class:`~texar.torch.modules.T5EncoderDecoder` class.

    """

    def setUp(self) -> None:
        self.batch_size = 2
        self.max_length = 3
        self.inputs = torch.zeros(
            self.batch_size, self.max_length, dtype=torch.long)

    @pretrained_test
    def test_model_loading(self):
        r"""Tests model loading functionality."""
        for pretrained_model_name in T5EncoderDecoder.available_checkpoints():
            if pretrained_model_name in ['T5-11B', 'T5-3B']:
                continue  # Too large to fit

            model = T5EncoderDecoder(
                pretrained_model_name=pretrained_model_name)
            _, _ = model(self.inputs)

    @pretrained_test
    def test_hparams(self):
        r"""Tests the priority of the architecture.
        """
        # case 1: set "pretrained_mode_name" by constructor argument
        hparams = {
            "pretrained_model_name": "T5-Small",
        }
        t5 = T5EncoderDecoder(pretrained_model_name="T5-Base",
                                 hparams=hparams)
        self.assertEqual(t5.hparams.encoder.num_blocks, 12)
        _, _ = t5(self.inputs)

        # case 2: set "pretrained_mode_name" by hparams
        hparams = {
            "pretrained_model_name": "T5-Small",
            "encoder": {
                "num_blocks": 16,
            }
        }
        t5 = T5EncoderDecoder(hparams=hparams)
        self.assertEqual(t5.hparams.encoder.num_blocks, 6)
        _, _ = t5(self.inputs)

        # case 3: set to None in both hparams and constructor argument
        hparams = {
            "pretrained_model_name": None,
            "encoder": {
                "num_blocks": 6,
            },
        }
        t5 = T5EncoderDecoder(hparams=hparams)
        self.assertEqual(t5.hparams.encoder.num_blocks, 6)
        _, _ = t5(self.inputs)

        # case 4: using default hparams
        encoder = T5EncoderDecoder()
        self.assertEqual(encoder.hparams.encoder.num_blocks, 6)
        _, _ = encoder(self.inputs)

    @pretrained_test
    def test_trainable_variables(self):
        r"""Tests the functionality of automatically collecting trainable
        variables.
        """
        # case 1: t5 small
        encoder = T5EncoderDecoder(pretrained_model_name="T5-Small")
        self.assertEqual(len(encoder.trainable_variables),
                         13 * 6 + 3 + 8 * 6 + 3)
        _, _ = encoder(self.inputs)

        # case 2: bert large
        hparams = {
            "pretrained_model_name": "T5-Base"
        }
        encoder = T5EncoderDecoder(hparams=hparams)
        self.assertEqual(len(encoder.trainable_variables),
                         13 * 12 + 3 + 8 * 12 + 3)
        _, _ = encoder(self.inputs)

    @pretrained_test
    def test_t5_eval(self):
        r"""Tests pre-trained model and check it generates
        same results everytime.
        """
        hparams = {
            "pretrained_model_name": 'T5-Small',
        }
        model = T5EncoderDecoder(hparams=hparams)
        model.eval()

        self.inputs = torch.from_numpy(
            numpy.asarray([[8774, 6, 82, 1782, 19, 5295]]))
        self.max_length = 6

        encoder_output, decoder_output = model(self.inputs)

        outputs_dim = model.hparams.encoder.dim
        self.assertEqual(
            decoder_output[0].shape,
            torch.Size([self.inputs.size()[0], self.max_length, outputs_dim]))

        self.assertEqual(
            encoder_output.shape,
            torch.Size([self.inputs.size()[0], self.max_length, outputs_dim]))

        # Check if these value are same consistently. If not, there is something
        # wrong with the pre-trained model.
        self.assertEqual(
            encoder_output.data[0][3][345].tolist(),
            -0.16204041242599487
        )
        self.assertLess(  # leave some margin for minor stochastic differences
            decoder_output[0].data[0][0][234].tolist() + 0.325570285320282,
            0.000001
        )

    def test_t5(self):
        r"""t5 test.
        """
        hparams = {
           "pretrained_model_name": None,
        }

        t5 = T5EncoderDecoder(hparams=hparams)

        inputs = torch.randint(32128, (self.batch_size, self.max_length))

        encoder_output, decoder_output = t5(inputs)

        outputs_dim = t5.output_size

        self.assertEqual(
            decoder_output[0].shape,
            torch.Size([self.batch_size, self.max_length, outputs_dim]))

        self.assertEqual(
            encoder_output.shape,
            torch.Size([self.batch_size, self.max_length, outputs_dim]))


if __name__ == "__main__":
    unittest.main()



# Copyright 2019 The Texar Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Executor metrics for generation tasks.
"""

import collections
import math
from typing import Counter, List, Sequence, Tuple

from texar.torch.run.metric.base_metric import StreamingMetric
from texar.torch.utils.types import MaybeList

__all__ = [
    "BLEU",
]


def _get_ngrams(segment: MaybeList[str],
                max_order: int) -> Counter[Tuple[str, ...]]:
    r"""Extracts all n-grams up to a given maximum order from an
    input segment.

    Args:
        segment: text segment from which n-grams will be extracted.
        max_order: maximum length in tokens of the n-grams returned
            by this methods.

    Returns:
        The Counter containing all n-grams upto :attr:`max_order`
        in segment with a count of how many times each n-gram occurred.
    """
    ngram_counts: Counter[Tuple[str, ...]] = collections.Counter()
    for order in range(1, max_order + 1):
        for i in range(0, len(segment) - order + 1):
            ngram = tuple(segment[i:i + order])
            ngram_counts[ngram] += 1
    return ngram_counts


class BLEU(StreamingMetric[MaybeList[str], float]):
    r"""The BLEU metric for evaluating translation tasks. BLEU stands for
    bilingual evaluation understudy, and measure the percentage of overlapping
    n-grams between the translation (hypothesis) and references.

    This metric also
    supports Smooth BLEU, computed following the method outlined in the paper:

        (Lin et al. 2004) ORANGE: a method for evaluating automatic evaluation
        metrics for machine translation.
        Chin-Yew Lin, Franz Josef Och. COLING 2004.

    BLEU is a :class:`~texar.torch.run.metric.StreamingMetric`, requires both
    predicted values and labels. BLEU values are :class:`float` numbers between
    0 and 100, with higher values being better.

    Args:
        max_order (int): Maximum n-gram order to use when computing BLEU score.
            Defaults to 4.
        lowercase (bool): Whether to lowercase all text before computing. If
            enabled, the metric is also known as "uncased BLEU". Defaults to
            `False`.
        smooth (bool): Whether or not to apply `(Lin et al. 2004)` smoothing.
            Defaults to `False`.
        brevity_penalty (bool): Whether to apply brevity penalty. Defaults to
            `True`.

    Keyword Args:
        pred_name (str): Name of the predicted value. This will be used as the
            key to the dictionary returned by the model.
        label_name (str): Name of the label. This will be used as the key to the
            batch object returned by the dataset.
    """
    reference_length: int
    hypothesis_length: int
    matches_by_order: List[int]
    possible_matches_by_order: List[int]

    def __init__(self, max_order: int = 4, lowercase: bool = False,
                 smooth: bool = False, brevity_penalty: bool = True,
                 *, pred_name: str, label_name: str):
        self.max_order = max_order
        self.lowercase = lowercase
        self.smooth = smooth
        self.brevity_penalty = brevity_penalty
        super().__init__(pred_name=pred_name, label_name=label_name)

    def reset(self) -> None:
        self.reference_length = 0
        self.hypothesis_length = 0
        self.matches_by_order = [0] * self.max_order
        self.possible_matches_by_order = [0] * self.max_order

    def add(self, predicted: Sequence[MaybeList[str]],
            labels: Sequence[MaybeList[str]]) -> None:
        for (reference, hypothesis) in zip(labels, predicted):
            if isinstance(reference, str):
                reference = reference.split()
            if self.lowercase:
                reference = [x.lower() for x in reference]
            reference_ngram_counts = _get_ngrams(reference, self.max_order)

            if isinstance(hypothesis, str):
                hypothesis = hypothesis.split()
            if self.lowercase:
                hypothesis = [x.lower() for x in hypothesis]
            hypothesis_ngram_counts = _get_ngrams(hypothesis, self.max_order)

            self.reference_length += len(reference)
            self.hypothesis_length += len(hypothesis)

            overlap = hypothesis_ngram_counts & reference_ngram_counts
            for ngram in overlap:
                self.matches_by_order[len(ngram) - 1] += overlap[ngram]
            for order in range(self.max_order):
                possible_matches = len(hypothesis) - order
                self.possible_matches_by_order[order] += possible_matches

    def value(self) -> float:
        if self.reference_length == 0:
            return 0.0
        precisions = []
        for i in range(self.max_order):
            if self.smooth:
                precision = ((self.matches_by_order[i] + 1.0) /
                             (self.possible_matches_by_order[i] + 1.0))
            elif self.possible_matches_by_order[i] > 0:
                precision = (self.matches_by_order[i] /
                             self.possible_matches_by_order[i])
            else:
                precision = 0.0
            precisions.append(precision)

        geo_mean = 0.0
        if min(precisions) > 0:
            p_log_sum = sum(math.log(p) for p in precisions) / self.max_order
            geo_mean = math.exp(p_log_sum)

        bp = 1.0
        if self.brevity_penalty:
            ratio = self.hypothesis_length / self.reference_length
            if ratio > 1.0:
                bp = 1.0
            elif abs(ratio) < 1e-8:
                bp = 0.0
            else:
                bp = math.exp(1 - 1.0 / ratio)

        bleu = geo_mean * bp
        return bleu * 100.0



# Copyright 2019 The Texar Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Unit tests for feed forward neural networks.
"""

import unittest

import torch

from texar.torch.modules.networks.networks import FeedForwardNetwork


class FeedForwardNetworkTest(unittest.TestCase):
    """Tests the class
    :class:`~texar.torch.modules.networks.networks.FeedForwardNetwork`.
    """

    def test_feedforward(self):
        """Tests feed-forward.
        """
        hparams = {
            "layers": [
                {
                    "type": "torch.nn.Linear",
                    "kwargs": {
                        "in_features": 32,
                        "out_features": 64
                    }
                },
                {
                    "type": "torch.nn.Linear",
                    "kwargs": {
                        "in_features": 64,
                        "out_features": 128
                    }
                }
            ]
        }

        nn = FeedForwardNetwork(hparams=hparams)

        self.assertEqual(len(nn.layers), len(hparams["layers"]))
        outputs = nn(torch.ones(64, 16, 32))
        self.assertEqual(len(nn.trainable_variables),
                         len(hparams["layers"]) * 2)
        self.assertEqual(outputs.size(-1), nn.output_size)


if __name__ == "__main__":
    unittest.main()



# Copyright 2019 The Texar Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Unit tests for embedders.
"""

import unittest

import numpy as np
import torch

from texar.torch.modules.embedders.embedders import WordEmbedder
from texar.torch.modules.embedders.position_embedders import (
    PositionEmbedder, SinusoidsPositionEmbedder)


class EmbedderTest(unittest.TestCase):
    """Tests parameterized embedder.
    """

    def _test_word_embedder(self, hparams):
        """Tests :class:`texar.torch.modules.WordEmbedder`.
        """
        embedder = WordEmbedder(
            vocab_size=100, hparams=hparams)

        inputs = torch.randint(embedder.vocab_size, (64, 16), dtype=torch.long)
        outputs = embedder(inputs)

        inputs_soft = torch.randn(
            (64, 16, embedder.vocab_size), dtype=torch.float32)
        outputs_soft = embedder(soft_ids=inputs_soft)

        if isinstance(embedder.dim, (list, tuple)):
            emb_dim = tuple(embedder.dim)
        else:
            emb_dim = (embedder.dim,)

        if isinstance(hparams["dim"], (list, tuple)):
            hparams_dim = tuple(hparams["dim"])
        else:
            hparams_dim = (hparams["dim"],)

        self.assertEqual(outputs.size(), (64, 16) + emb_dim)
        self.assertEqual(outputs.size(-1), embedder.output_size)
        self.assertEqual(outputs_soft.size(), (64, 16) + emb_dim)
        self.assertEqual(outputs_soft.size(-1), embedder.output_size)
        self.assertEqual(emb_dim, hparams_dim)
        self.assertEqual(embedder.vocab_size, 100)
        self.assertEqual(outputs.size(), (64, 16) + emb_dim)
        self.assertEqual(outputs_soft.size(), (64, 16) + emb_dim)

    def _test_position_embedder(self, hparams):
        """Tests :class:`texar.torch.modules.PositionEmbedder`.
        """
        pos_size = 100
        embedder = PositionEmbedder(
            position_size=pos_size, hparams=hparams)
        inputs = torch.randint(embedder.num_embeds, (64, 16), dtype=torch.long)
        outputs = embedder(inputs)

        if isinstance(embedder.dim, (list, tuple)):
            emb_dim = tuple(embedder.dim)
        else:
            emb_dim = (embedder.dim,)

        if isinstance(hparams["dim"], (list, tuple)):
            hparams_dim = tuple(hparams["dim"])
        else:
            hparams_dim = (hparams["dim"],)

        self.assertEqual(outputs.size(), (64, 16) + emb_dim)
        self.assertEqual(outputs.size(-1), embedder.output_size)
        self.assertEqual(emb_dim, hparams_dim)
        self.assertEqual(embedder.position_size, 100)
        seq_length = torch.empty(64).uniform_(0, pos_size).long()
        _ = embedder(sequence_length=seq_length)

    def test_sinusoids_position_embedder(self):
        """Tests :class:`texar.torch.modules.SinusoidsPositionEmbedder`.
        """
        position_size = 64
        input_size = (23, 18)
        hparams = {'dim': 513}  # use odd dimension to ensure padding correct
        embedder = SinusoidsPositionEmbedder(position_size, hparams=hparams)
        inputs = torch.randint(position_size - 1, input_size)
        outputs = embedder(inputs)
        self.assertEqual(outputs.size(), input_size + (hparams['dim'],))
        self.assertEqual(outputs.size(-1), embedder.output_size)
        sequence_length = torch.randint(high=position_size + 1, size=(23,))
        outputs_ = embedder(sequence_length=sequence_length)
        self.assertEqual(outputs_.size(),
                         (23, max(sequence_length).item()) + (hparams['dim'],))
        self.assertEqual(outputs_.size(-1), embedder.output_size)

        embedder_no_cache = SinusoidsPositionEmbedder(
            None, hparams={**hparams, 'cache_embeddings': False})
        wide_inputs = torch.randint(
            -position_size, position_size * 2, input_size)
        wide_outputs = embedder_no_cache(wide_inputs)
        self.assertEqual(wide_outputs.size(), input_size + (hparams['dim'],))
        no_cache_outputs = embedder_no_cache(inputs)
        np.testing.assert_array_equal(outputs, no_cache_outputs)

    def test_embedder(self):
        """Tests various embedders.
        """
        test_dims = [
            14,
            [14],
            [14, 10],
        ]
        test_hparams = [
            # no dropout
            {"dropout_rate": 0},
            # dropout with default strategy
            {"dropout_rate": 0.3},
            # dropout with different strategies
            {"dropout_rate": 0.3, "dropout_strategy": "item"},
            {"dropout_rate": 0.3, "dropout_strategy": "item_type"},
        ]

        for base_hparams in test_hparams:
            for dim in test_dims:
                hparams = base_hparams.copy()
                hparams['dim'] = dim
                self._test_word_embedder(hparams)
                self._test_position_embedder(hparams)

    def test_embedder_multi_calls(self):
        """Tests embedders called by multiple times.
        """
        hparams = {"dim": 26, "dropout_rate": 0.3,
                   "dropout_strategy": "item"}
        embedder = WordEmbedder(
            vocab_size=100, hparams=hparams)
        inputs = torch.randint(embedder.vocab_size, (64, 16), dtype=torch.long)
        outputs = embedder(inputs)

        if isinstance(embedder.dim, (list, tuple)):
            emb_dim = tuple(embedder.dim)
        else:
            emb_dim = (embedder.dim,)
        self.assertEqual(outputs.size(), (64, 16) + emb_dim)

        # Call with inputs in a different shape
        inputs = torch.randint(
            embedder.vocab_size, (64, 10, 20), dtype=torch.long)
        outputs = embedder(inputs)

        self.assertEqual(outputs.size(), (64, 10, 20) + emb_dim)

    def test_word_embedder_soft_ids(self):
        """Tests the correctness of using soft ids.
        """
        init_value = np.expand_dims(np.arange(5), 1)
        embedder = WordEmbedder(init_value=init_value)

        ids = torch.tensor([3])
        soft_ids = torch.tensor([0, 0, 0, 1, 0], dtype=torch.float)

        outputs = embedder(ids=ids)
        soft_outputs = embedder(soft_ids=soft_ids)
        self.assertEqual(outputs, soft_outputs)

    def test_word_embedder_trainable(self):
        """Tests freezing the embedding parameters.
        """
        init_value = np.expand_dims(np.arange(5), 1)

        embedder = WordEmbedder(init_value=init_value,
                                hparams={"trainable": False})
        self.assertEqual(len(embedder.trainable_variables), 0)

        embedder = WordEmbedder(init_value=init_value)
        self.assertEqual(len(embedder.trainable_variables), 1)

    def test_position_embedder_trainable(self):
        """Tests freezing the embedding parameters.
        """
        pos_size = 100

        embedder = PositionEmbedder(
            position_size=pos_size, hparams={"trainable": False})
        self.assertEqual(len(embedder.trainable_variables), 0)

        embedder = PositionEmbedder(position_size=pos_size)
        self.assertEqual(len(embedder.trainable_variables), 1)


if __name__ == "__main__":
    unittest.main()



# Copyright 2019 The Texar Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import ast
import math
from typing import Dict, Optional, Tuple

import torch
from torch.nn import functional as F
from torch import nn

from texar.torch.core import layers
from texar.torch.module_base import ModuleBase
from texar.torch.modules.encoders.multihead_attention import LayerCache
from texar.torch.utils.types import MaybeList

__all__ = [
    "T5LayerNorm",
    "MultiheadRPRAttention",
    "read_t5_gin_config_file"
]

IMPORTANT_PARAMS = ('d_ff',
                    'd_kv',
                    'd_model',
                    'dropout',
                    'num_heads',
                    'num_layers',
                    'inputs_length'
                    )


def read_t5_gin_config_file(config_file_path: str) -> Dict:
    r"""Simple helper function to read a gin file
    and get hyperparameters for T5.

    Args:
        config_file_path: path of config.gin file as a string.

    Returns:
        A dictionary with important parameters for loading T5.

    """
    config = {}

    with open(config_file_path, 'r') as gin_file:
        for line in gin_file:
            if line.startswith(IMPORTANT_PARAMS):
                assignment = line.strip().split()
                assert len(assignment) == 3
                arg_name, _, value = assignment
                config[arg_name] = ast.literal_eval(value)

    return config


class T5LayerNorm(nn.Module):
    r""" Custom LayerNorm for T5 with no mean subtraction and no bias.
    """

    def __init__(self, input_size: int, eps: float = 1e-5):
        super().__init__()

        self.w = nn.Parameter(torch.ones(input_size))
        self.eps = eps

    def forward(self,  # type: ignore
                x: torch.Tensor):
        x = x / torch.sqrt(x.pow(2).mean(-1, keepdim=True) + self.eps)
        return self.w * x


class MultiheadRPRAttention(ModuleBase):
    r"""
    A variation of MultiheadAttention introduced by

    Peter Shaw, Jakob Uszkoreit, and Ashish Vaswani. "Self-attention with
    relative position representations".  https://arxiv.org/pdf/1803.02155.pdf

    wherein it uses an alternative means of encoding positional information
    of an input sequence by learning an embedding which is stored in the
    first layer of Encoder/Decoder Stack but is shared amongst all attention
    layers in the stack.

    Args:
        input_size (int): Number of hidden dimension

        hparams (dict or HParams, optional): Hyperparameters. Missing
            hyperparameters will be set to default values. See
            :meth:`default_hparams` for the hyperparameter structure and
            default values.

        stores_relative_position (bool): If the instance stores the learned
        realtive position embeddings.

    .. document private functions
    """

    def __init__(self, input_size: int, hparams=None,
                 stores_relative_position: bool = False):
        super().__init__(hparams=hparams)

        use_bias = self._hparams.use_bias
        self.is_decoder = self._hparams.is_decoder
        self.stores_rpr = stores_relative_position
        self.relative_attention_num_buckets = \
            self._hparams.relative_attention_num_buckets

        self.Q_dense = nn.Linear(input_size, self._hparams.num_units,
                                 bias=use_bias)
        self.K_dense = nn.Linear(input_size, self._hparams.num_units,
                                 bias=use_bias)
        self.V_dense = nn.Linear(input_size, self._hparams.num_units,
                                 bias=use_bias)
        self.O_dense = nn.Linear(self._hparams.num_units,
                                 self._hparams.output_dim, bias=use_bias)

        if self.stores_rpr:
            self.relative_attention_bias = nn.Embedding(
                self.relative_attention_num_buckets, self._hparams.num_heads)

        if self._hparams.initializer:
            # TODO(haoransh): we may define kernel_initializer and bias
            #  initializer seperately
            initialize = layers.get_initializer(self._hparams.initializer)
            assert initialize is not None
            for name, param in self.named_parameters():
                if name.split('.')[-1] == 'weight':
                    print('name:{}'.format(name))
                    initialize(param)

    @staticmethod
    def default_hparams():
        r"""Returns a dictionary of hyperparameters with default values.

        .. code-block:: python

            {
                "initializer": None,
                'num_heads': 8,
                'output_dim': 512,
                'num_units': 512,
                'dropout_rate': 0.1,
                'use_bias': False,
                "name": "multihead_attention",
                "is_decoder": False,
                "relative_attention_num_buckets": 32
            }

        Here:

        `"initializer"`: dict, optional
            Hyperparameters of the default initializer that initializes
            variables created in this module.
            See :func:`~texar.torch.core.get_initializer` for details.

        `"num_heads"`: int
            Number of heads for attention calculation.

        `"output_dim"`: int
            Output dimension of the returned tensor.

        `"num_units"`: int
            Hidden dimension of the unsplit attention space.
            Should be divisible by `"num_heads"`.

        `"dropout_rate"`: float
            Dropout rate in the attention.

        `"use_bias"`: bool
            Use bias when projecting the key, value and query.

        `"name"`: str
            Name of the module.

        `"is_decoder"`: bool
            To pass in if the attention is for a encoder or decoder block.

        `"name"`: relative_attention_num_buckets
            If the Attention mechnanism needs to use relative positional
            attention bias, then this hparam stores the relative attention
            num buckets.
        """
        return {
            'initializer': None,
            'num_heads': 8,
            'output_dim': 512,
            'num_units': 512,
            'dropout_rate': 0.1,
            'use_bias': False,
            'name': 'multihead_attention_rpr',
            'is_decoder': False,
            'relative_attention_num_buckets': 32
        }

    @staticmethod
    def _relative_position_bucket(relative_position,
                                  bidirectional=True,
                                  num_buckets=32,
                                  max_distance=128):
        """
        Adapted from Mesh Tensorflow:
        https://github.com/tensorflow/mesh/blob/master/
        mesh_tensorflow/transformer/transformer_layers.py#L595

        Translate relative position to a bucket number for relative attention.

        The relative position is defined as memory_position - query_position,
        i.e. the distance in tokens from the attending position to the
        attended-to position.  If bidirectional=False, then positive relative
        positions are invalid.

        We use smaller buckets for small absolute relative_position and
        larger buckets for larger absolute relative_positions.

        All relative positions >=max_distance map to the same bucket.
        All relative positions <=-max_distance map to the same bucket.

        This should allow for more graceful generalization to longer
        sequences than the model has been trained on.

        Args:
            relative_position: an int32 Tensor
            bidirectional: a boolean - whether the attention is bidirectional
            num_buckets: an integer
            max_distance: an integer
        Returns:
            a Tensor with the same shape as relative_position, containing int32
            values in the range [0, num_buckets)
        """
        ret = 0
        n = -relative_position
        if bidirectional:
            num_buckets //= 2
            # mtf.to_int32(mtf.less(n, 0)) * num_buckets
            ret += (n < 0).to(torch.long) * num_buckets
            n = torch.abs(n)
        else:
            n = torch.max(n, torch.zeros_like(n))
        # now n is in the range [0, inf)

        # half of the buckets are for exact increments in positions
        max_exact = num_buckets // 2
        is_small = (n < max_exact)

        # The other half of the buckets are for logarithmically bigger bins
        # in positions up to max_distance
        val_if_large = max_exact + (
                torch.log(n.float() / max_exact)
                / math.log(max_distance / max_exact) * (
                            num_buckets - max_exact)).to(torch.long)
        val_if_large = torch.min(val_if_large,
                                 torch.full_like(val_if_large, num_buckets - 1))

        ret += torch.where(is_small, n, val_if_large)
        return ret

    def compute_bias(self, qlen, klen):
        """ Compute binned relative position bias """
        context_position = torch.arange(qlen, dtype=torch.long)[:, None]
        memory_position = torch.arange(klen, dtype=torch.long)[None, :]

        relative_position = memory_position - context_position
        #  [length_query, length_key]

        rp_bucket = self._relative_position_bucket(
            relative_position,
            bidirectional=not self.is_decoder,
            num_buckets=self.relative_attention_num_buckets)
        # [length_query, length_key]

        values = self.relative_attention_bias(rp_bucket)
        # [length_query, length_key, num_heads]

        values = values.permute([2, 0, 1]).unsqueeze(0)
        return values  # [1, num_heads, length_query, length_key]

    def forward(self,  # type: ignore
                queries: torch.Tensor,
                memory: torch.Tensor,
                memory_attention_bias: torch.Tensor,
                cache: Optional[LayerCache] = None,
                position_bias: Optional[torch.Tensor] = None
                ) \
            -> Tuple[torch.Tensor, torch.Tensor]:
        r"""Encodes the inputs.

        Args:
            queries: A 3D tensor with shape of
                ``[batch, length_query, depth_query]``.
            memory: A 4D tensor with shape of
                ``[batch, length_key, depth_key]``.
            memory_attention_bias: A 4D tensor with shape of
                ``[batch, length_key, num_units]``.
            cache: Memory cache only when inferring the sentence from scratch.
            position_bias: A 4D Tensor with shape of
                ``[batch, num_heads, length_query, length_query)]``.

        Returns:
            A tuple of 2 tensors, first, a tensor of shape
            ``[batch_size, max_time, dim]`` containing the encoded vectors
            and second a tensor of shape
            [batch, num_heads, length_query, length_query)] for sharing the
            position bias
        """
        length_query = queries.size(1)
        if memory is None:
            length_key = length_query  # Self-Attention
        else:
            length_key = memory.size(1)

        num_heads = self._hparams.num_heads
        num_units = self._hparams.num_units

        if num_units % num_heads != 0:
            raise ValueError(
                f"Value depth ({num_units}) must be divisible by "
                f"the number of attention heads ({num_heads}).")

        def _update_and_return(layer: nn.Module, key: str):
            if memory is None:
                # Self Attention
                out = layer(queries)

                if cache is not None:
                    # decoder self attention when dynamic decoding
                    res: MaybeList[torch.Tensor] = cache[key]
                    if isinstance(res, list):
                        # inference-like decoding
                        res.append(out.squeeze(1))
                        out = torch.stack(res, dim=1)
                    else:
                        # normal decoding
                        res = torch.cat([res, out], dim=1)
                        out = res
                    cache[key] = res

            else:
                # encoder decoder attention
                if cache is not None:
                    res: MaybeList[torch.Tensor] = cache[  # type: ignore
                        key]  # type: ignore
                    if isinstance(res, list):
                        # inference-like decoding
                        if len(res) == 0:
                            out = layer(memory)
                        else:
                            out = torch.stack(res, dim=1)
                    else:
                        # normal decoding
                        if res.size(1) == 0:
                            out = layer(memory)
                        else:
                            out = res
                else:
                    out = layer(memory)

            return out

        Q = self.Q_dense(queries)
        K = _update_and_return(self.K_dense, 'keys')
        V = _update_and_return(self.V_dense, 'values')

        Q_ = self._split_heads(Q)
        K_ = self._split_heads(K)
        V_ = self._split_heads(V)

        # All of the above [batch_size, num_heads, seq_length, memory_depth]
        # Q_ *= key_depth_per_head ** -0.5  # T5 does not scale

        logits = torch.einsum('bnqd,bnkd->bnqk', Q_, K_)  # type: ignore

        if position_bias is None:
            # Must compute bias using embedding stored.
            # Check if self.stores_rpr is True
            if not self.stores_rpr:
                raise ValueError("Layer must store embedding weights since"
                                 "relative bias not provided")
            position_bias = self.compute_bias(length_query, length_key)

            if memory_attention_bias is not None:
                memory_attention_bias = memory_attention_bias.to(
                    device=logits.device)
                position_bias = position_bias + memory_attention_bias

        logits += position_bias
        weights = torch.softmax(logits, dim=-1)
        weights = F.dropout(weights, self._hparams.dropout_rate,
                            self.training)
        outputs = torch.matmul(weights, V_)

        outputs = self._combine_heads(outputs)
        outputs = self.O_dense(outputs)
        # (batch_size, length_query, output_dim)

        return outputs, position_bias

    def _split_heads(self, x: torch.Tensor) -> torch.Tensor:
        r"""Split channels (dimension 2) into multiple heads,
        becomes dimension 1). Must ensure ``x.shape[-1]`` can be
        divided by num_heads.
        """
        depth = x.size(-1)
        split_x = torch.reshape(x, (
            x.size(0), x.size(1),
            self._hparams.num_heads, depth // self._hparams.num_heads))
        return split_x.permute((0, 2, 1, 3))

    def _combine_heads(self, x: torch.Tensor) -> torch.Tensor:
        r"""

        Args:
            x: A Tensor of shape ``[batch, num_heads, seq_len, dim]``
        Returns:
            A Tensor of shape ``[batch, seq_len, num_heads * dim]``
        """
        t = x.permute((0, 2, 1, 3))  # [batch, seq_len, num_heads, dim]
        num_heads, dim = t.size()[-2:]
        assert num_heads == self._hparams.num_heads
        return torch.reshape(t, (t.size(0), t.size(1), num_heads * dim))

    @property
    def output_size(self):
        r"""The feature size of :meth:`forward` output.
        """
        return self._hparams.output_dim



# Copyright 2019 The Texar Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Unit tests for pg losses.
"""

import unittest

import torch

from texar.torch.losses.pg_losses import pg_loss_with_logits
from texar.torch.utils.shapes import get_rank


class PGLossesTest(unittest.TestCase):
    """Tests pg losses
    """

    def setUp(self):
        self._batch_size = 64
        self._max_time = 16
        self._d1 = 3  # use smaller values to speedup testing
        self._d2 = 4
        self._d3 = 5
        self._num_classes = 10
        self._actions_batch = torch.ones(
            self._batch_size, self._max_time, self._d1, self._d2, self._d3,
            dtype=torch.int64)
        self._logits_batch = torch.rand(
            self._batch_size, self._max_time, self._d1, self._d2, self._d3,
            self._num_classes)
        self._advantages_batch = torch.rand(
            self._batch_size, self._max_time, self._d1, self._d2, self._d3)
        self._actions_no_batch = torch.ones(
            self._max_time, self._d1, self._d2, self._d3, dtype=torch.int64)
        self._logits_no_batch = torch.rand(
            self._max_time, self._d1, self._d2, self._d3, self._num_classes)
        self._advantages_no_batch = torch.rand(
            self._max_time, self._d1, self._d2, self._d3)
        self._sequence_length = torch.randint(
            high=self._max_time, size=(self._batch_size,))

    def _test_sequence_loss(self, loss_fn, actions, logits, advantages, batched,
                            sequence_length):
        loss = loss_fn(actions, logits, advantages, batched=batched,
                       sequence_length=sequence_length)
        rank = get_rank(loss)
        self.assertEqual(rank, 0)

        loss = loss_fn(actions, logits, advantages, batched=batched,
                       sequence_length=sequence_length,
                       sum_over_timesteps=False)
        rank = get_rank(loss)
        self.assertEqual(rank, 1)
        self.assertEqual(loss.shape, torch.Size([self._max_time]))

        loss = loss_fn(actions, logits, advantages, batched=batched,
                       sequence_length=sequence_length,
                       sum_over_timesteps=False,
                       average_across_timesteps=True,
                       average_across_batch=False)
        rank = get_rank(loss)
        if batched:
            self.assertEqual(rank, 1)
            self.assertEqual(loss.shape, torch.Size([self._batch_size]))
        else:
            self.assertEqual(rank, 0)

        loss = loss_fn(actions, logits, advantages, batched=batched,
                       sequence_length=sequence_length,
                       sum_over_timesteps=False,
                       average_across_batch=False)
        rank = get_rank(loss)
        if batched:
            self.assertEqual(rank, 2)
            self.assertEqual(loss.shape,
                             torch.Size([self._batch_size, self._max_time]))
        else:
            self.assertEqual(rank, 1)
            self.assertEqual(loss.shape,
                             torch.Size([self._max_time]))

        sequence_length_time = torch.randint(
            high=self._batch_size, size=(self._max_time,))
        loss = loss_fn(actions, logits, advantages, batched=batched,
                       sequence_length=sequence_length_time,
                       sum_over_timesteps=False,
                       average_across_batch=False,
                       time_major=True)
        if batched:
            self.assertEqual(loss.shape, torch.Size([self._batch_size,
                                                     self._max_time]))
        else:
            self.assertEqual(loss.shape, torch.Size([self._max_time]))

    def test_pg_loss_with_logits(self):
        """Tests `texar.torch.losses.pg_loss_with_logits`.
        """
        self._test_sequence_loss(
            pg_loss_with_logits,
            self._actions_batch, self._logits_batch,
            self._advantages_batch, True, self._sequence_length)

        self._test_sequence_loss(
            pg_loss_with_logits,
            self._actions_no_batch, self._logits_no_batch,
            self._advantages_no_batch, False, self._sequence_length)


if __name__ == "__main__":
    unittest.main()



# Copyright 2019 The Texar Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Unit tests for attention mechanism.
"""

import unittest

import numpy as np
import torch

from texar.torch.core.attention_mechanism import *


class AttentionMechanismTest(unittest.TestCase):
    r"""Tests attention mechanism.
    """

    def setUp(self):
        self._batch_size = 8
        self._max_time = 16
        self._encoder_output_size = 64
        self._attention_dim = 256
        self._memory = torch.rand(
            self._batch_size, self._max_time, self._encoder_output_size
        )
        self._memory_sequence_length = torch.tensor(
            np.random.randint(self._max_time, size=[self._batch_size]) + 1
        )
        self._attention_state = torch.rand(self._batch_size, self._max_time)

    def test_LuongAttention(self):
        r"""Tests `LuongAttention`
        """
        # Case 1
        attention_mechanism = LuongAttention(
            num_units=self._attention_dim,
            encoder_output_size=self._encoder_output_size)

        cell_output = torch.rand(self._batch_size, self._attention_dim)

        attention, alignments, next_attention_state = \
            compute_attention(
                attention_mechanism=attention_mechanism,
                cell_output=cell_output,
                attention_state=self._attention_state,
                memory=self._memory,
                attention_layer=None,
                memory_sequence_length=self._memory_sequence_length)

        self.assertEqual(attention.shape, torch.Size(
            [self._batch_size, self._encoder_output_size]))
        self.assertEqual(alignments.shape, torch.Size(
            [self._batch_size, self._max_time]))
        self.assertEqual(next_attention_state.shape, torch.Size(
            [self._batch_size, self._max_time]))
        self.assertEqual(len(attention_mechanism.trainable_variables), 1)

        # Case 2
        attention_mechanism = LuongAttention(
            num_units=self._attention_dim,
            encoder_output_size=self._encoder_output_size,
            scale=True)

        cell_output = torch.rand(self._batch_size, self._attention_dim)

        attention, alignments, next_attention_state = \
            compute_attention(
                attention_mechanism=attention_mechanism,
                cell_output=cell_output,
                attention_state=self._attention_state,
                memory=self._memory,
                attention_layer=None,
                memory_sequence_length=self._memory_sequence_length)

        self.assertEqual(attention.shape, torch.Size(
            [self._batch_size, self._encoder_output_size]))
        self.assertEqual(alignments.shape, torch.Size(
            [self._batch_size, self._max_time]))
        self.assertEqual(next_attention_state.shape, torch.Size(
            [self._batch_size, self._max_time]))
        self.assertEqual(len(attention_mechanism.trainable_variables), 2)

    def test_BahdanauAttention(self):
        r"""Tests BahdanauAttention
        """
        # Case 1
        attention_mechanism = BahdanauAttention(
            num_units=self._attention_dim,
            decoder_output_size=128,
            encoder_output_size=self._encoder_output_size)

        cell_output = torch.rand(self._batch_size, 128)

        attention, alignments, next_attention_state = \
            compute_attention(
                attention_mechanism=attention_mechanism,
                cell_output=cell_output,
                attention_state=self._attention_state,
                memory=self._memory,
                attention_layer=None,
                memory_sequence_length=self._memory_sequence_length)

        self.assertEqual(attention.shape, torch.Size(
            [self._batch_size, self._encoder_output_size]))
        self.assertEqual(alignments.shape, torch.Size(
            [self._batch_size, self._max_time]))
        self.assertEqual(next_attention_state.shape, torch.Size(
            [self._batch_size, self._max_time]))
        self.assertEqual(len(attention_mechanism.trainable_variables), 3)

        # Case 2
        attention_mechanism = BahdanauAttention(
            num_units=self._attention_dim,
            decoder_output_size=128,
            encoder_output_size=self._encoder_output_size,
            normalize=True)

        cell_output = torch.rand(self._batch_size, 128)

        attention, alignments, next_attention_state = \
            compute_attention(
                attention_mechanism=attention_mechanism,
                cell_output=cell_output,
                attention_state=self._attention_state,
                memory=self._memory,
                attention_layer=None,
                memory_sequence_length=self._memory_sequence_length)

        self.assertEqual(attention.shape, torch.Size(
            [self._batch_size, self._encoder_output_size]))
        self.assertEqual(alignments.shape, torch.Size(
            [self._batch_size, self._max_time]))
        self.assertEqual(next_attention_state.shape, torch.Size(
            [self._batch_size, self._max_time]))
        self.assertEqual(len(attention_mechanism.trainable_variables), 5)

    def test_LuongMonotonicAttention(self):
        r"""Tests LuongMonotonicAttention
        """
        # Case 1
        attention_mechanism = LuongMonotonicAttention(
            num_units=self._attention_dim,
            encoder_output_size=self._encoder_output_size)

        cell_output = torch.rand(self._batch_size, self._attention_dim)

        attention, alignments, next_attention_state = \
            compute_attention(
                attention_mechanism=attention_mechanism,
                cell_output=cell_output,
                attention_state=self._attention_state,
                memory=self._memory,
                attention_layer=None,
                memory_sequence_length=self._memory_sequence_length)

        self.assertEqual(attention.shape, torch.Size(
            [self._batch_size, self._encoder_output_size]))
        self.assertEqual(alignments.shape, torch.Size(
            [self._batch_size, self._max_time]))
        self.assertEqual(next_attention_state.shape, torch.Size(
            [self._batch_size, self._max_time]))
        self.assertEqual(len(attention_mechanism.trainable_variables), 2)

        # Case 2
        attention_mechanism = LuongMonotonicAttention(
            num_units=self._attention_dim,
            encoder_output_size=self._encoder_output_size,
            scale=True)

        cell_output = torch.rand(self._batch_size, self._attention_dim)

        attention, alignments, next_attention_state = \
            compute_attention(
                attention_mechanism=attention_mechanism,
                cell_output=cell_output,
                attention_state=self._attention_state,
                memory=self._memory,
                attention_layer=None,
                memory_sequence_length=self._memory_sequence_length)

        self.assertEqual(attention.shape, torch.Size(
            [self._batch_size, self._encoder_output_size]))
        self.assertEqual(alignments.shape, torch.Size(
            [self._batch_size, self._max_time]))
        self.assertEqual(next_attention_state.shape, torch.Size(
            [self._batch_size, self._max_time]))
        self.assertEqual(len(attention_mechanism.trainable_variables), 3)

    def test_BahdanauMonotonicAttention(self):
        r"""Tests BahdanauMonotonicAttention
        """
        # Case 1
        attention_mechanism = BahdanauMonotonicAttention(
            num_units=self._attention_dim,
            decoder_output_size=128,
            encoder_output_size=self._encoder_output_size)

        cell_output = torch.rand(self._batch_size, 128)

        attention, alignments, next_attention_state = \
            compute_attention(
                attention_mechanism=attention_mechanism,
                cell_output=cell_output,
                attention_state=self._attention_state,
                memory=self._memory,
                attention_layer=None,
                memory_sequence_length=self._memory_sequence_length)

        self.assertEqual(attention.shape, torch.Size(
            [self._batch_size, self._encoder_output_size]))
        self.assertEqual(alignments.shape, torch.Size(
            [self._batch_size, self._max_time]))
        self.assertEqual(next_attention_state.shape, torch.Size(
            [self._batch_size, self._max_time]))
        self.assertEqual(len(attention_mechanism.trainable_variables), 4)

        # Case 2
        attention_mechanism = BahdanauMonotonicAttention(
            num_units=self._attention_dim,
            decoder_output_size=128,
            encoder_output_size=self._encoder_output_size,
            normalize=True)

        cell_output = torch.rand(self._batch_size, 128)

        attention, alignments, next_attention_state = \
            compute_attention(
                attention_mechanism=attention_mechanism,
                cell_output=cell_output,
                attention_state=self._attention_state,
                memory=self._memory,
                attention_layer=None,
                memory_sequence_length=self._memory_sequence_length)

        self.assertEqual(attention.shape, torch.Size(
            [self._batch_size, self._encoder_output_size]))
        self.assertEqual(alignments.shape, torch.Size(
            [self._batch_size, self._max_time]))
        self.assertEqual(next_attention_state.shape, torch.Size(
            [self._batch_size, self._max_time]))
        self.assertEqual(len(attention_mechanism.trainable_variables), 6)


if __name__ == "__main__":
    unittest.main()



# Copyright 2019 The Texar Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""RNN helpers for PyTorch models."""

from typing import List, Optional, Tuple, TypeVar, Union

import torch

from texar.torch.core.cell_wrappers import RNNCellBase
from texar.torch.utils.shapes import mask_sequences
from texar.torch.utils.utils import map_structure, map_structure_zip, no_map

__all__ = [
    "reverse_sequence",
    "dynamic_rnn",
    "bidirectional_dynamic_rnn",
]

State = TypeVar('State')


def reverse_sequence(inputs: torch.Tensor,
                     seq_lengths: Union[torch.LongTensor, List[int]],
                     time_major: bool) -> torch.Tensor:
    r"""Reverses variable length slices.

    This op first slices input along the dimension batch_axis, and for each
    slice i, reverses the first seq_lengths[i] elements along the dimension
    seq_axis.

    The elements of seq_lengths must obey seq_lengths[i] <=
    input.dims[seq_dim], and seq_lengths must be a vector of length
    input.dims[batch_dim].

    The output slice i along dimension batch_axis is then given by input slice
    i, with the first seq_lengths[i] slices along dimension seq_axis reversed.

    Args:
        inputs: A Tensor. The input to reverse.
        seq_lengths: A Tensor. Must be one of the following types: int32,
            int64. 1-D with length input.dims(batch_dim) and
            max(seq_lengths) <= input.dims(seq_dim)
        time_major: The shape format of the ``inputs`` and ``outputs`` Tensors.
            If true, these ``Tensors`` must be shaped
            ``[max_time, batch_size, depth]``. If false, these ``Tensors`` must
            be shaped ``[batch_size, max_time, depth]``.
            Using ``time_major = True`` is a bit more efficient because it
            avoids transposes at the beginning and end of the RNN calculation.
            However, most TensorFlow data is batch-major, so by
            default this functionb accepts input and emits output
            in batch-major form.

    Returns:
        A ``Tensor``. Has the same type as input.
    """
    if time_major:
        inputs = inputs.permute(1, 0, 2)

    batch_size = inputs.shape[0]

    outputs = inputs.clone()
    for i in range(batch_size):
        outputs[i][0:seq_lengths[i]] = torch.flip(
            inputs[i][0:seq_lengths[i]], dims=(0,))
    if time_major:
        outputs = outputs.permute(1, 0, 2)

    return outputs


def bidirectional_dynamic_rnn(
        cell_fw: RNNCellBase[State],
        cell_bw: RNNCellBase[State],
        inputs: torch.Tensor,
        sequence_length: Optional[Union[torch.LongTensor, List[int]]] = None,
        initial_state_fw: Optional[State] = None,
        initial_state_bw: Optional[State] = None,
        time_major: bool = False) -> Tuple[Tuple[torch.Tensor, torch.Tensor],
                                           Tuple[State, State]]:
    r"""Creates a dynamic version of bidirectional recurrent neural network.

    Takes input and builds independent forward and backward RNNs. The
    input_size of forward and backward cell must match. The initial state for
    both directions is zero by default (but can be set optionally) and no
    intermediate states are ever returned -- the network is fully unrolled
    for the given (passed in) length(s) of the sequence(s) or completely
    unrolled if length(s) is not given.

    Args:
        cell_fw: An instance of RNNCell, to be used for forward direction.
        cell_bw: An instance of RNNCell, to be used for backward direction.
        inputs: The RNN inputs.
            If time_major == False (default), this must be a tensor of shape:
            ``[batch_size, max_time, ...]``, or a nested tuple of such elements.
            If time_major == True, this must be a tensor of shape:
            ``[max_time, batch_size, ...]``, or a nested tuple of such elements.
        sequence_length: (optional) An int32/int64 tensor, size
            ``[batch_size]``, containing the actual lengths for each of the
            sequences in
            the batch. If not provided, all batch entries are assumed
            to be full sequences; and time reversal is applied from time
            ``0`` to ``max_time`` for each sequence.
        initial_state_fw: (optional) An initial state for the forward RNN.
            This must be a tensor of appropriate type and shape
            ``[batch_size, cell_fw.state_size]``.
            If ``cell_fw.state_size`` is a tuple, this should be a tuple of
            tensors having shapes ``[batch_size, s]``
            for ``s`` in ``cell_fw.state_size``.
        initial_state_bw: (optional) Same as for ``initial_state_fw``, but using
            the corresponding properties of ``cell_bw``.
        time_major: The shape format of the ``inputs`` and ``outputs`` Tensors.
            If true, these ``Tensors`` must be shaped
            ``[max_time, batch_size, depth]``.
            If false, these ``Tensors`` must be shaped
            ``[batch_size, max_time, depth]``.
            Using ``time_major = True`` is a bit more efficient because it
            avoids transposes at the beginning and end of the RNN calculation.
            However, most TensorFlow data is batch-major, so by
            default this function accepts input and emits output
            in batch-major form.

    Returns:
        A tuple (outputs, output_states) where:

        outputs: A tuple (output_fw, output_bw) containing the forward and
            the backward rnn output ``Tensor``.
            If time_major == False (default),
                output_fw will be a ``Tensor`` shaped:
                ``[batch_size, max_time, cell_fw.output_size]``
                and output_bw will be a ``Tensor`` shaped:
                ``[batch_size, max_time, cell_bw.output_size]``.
            If time_major == True,
                output_fw will be a ``Tensor`` shaped:
                ``[max_time, batch_size, cell_fw.output_size]``
                and output_bw will be a ``Tensor`` shaped:
                ``[max_time, batch_size, cell_bw.output_size]``.
            It returns a tuple instead of a single concatenated ``Tensor``,
            unlike in the ``bidirectional_rnn``. If the concatenated
            one is preferred, the forward and backward outputs can
            be concatenated as ``tf.concat(outputs, 2)``.
        output_states: A tuple (output_state_fw, output_state_bw) containing
            the forward and the backward final states of bidirectional rnn.
    """
    output_fw, output_state_fw = dynamic_rnn(cell=cell_fw,
                                             inputs=inputs,
                                             sequence_length=sequence_length,
                                             initial_state=initial_state_fw,
                                             time_major=time_major)
    if time_major:
        time_steps = inputs.shape[0]
        batch_size = inputs.shape[1]
    else:
        time_steps = inputs.shape[1]
        batch_size = inputs.shape[0]

    if sequence_length is None:
        sequence_length = torch.tensor([time_steps] * batch_size,
                                       dtype=torch.int32,
                                       device=inputs.device)

    # Backward direction
    inputs_reverse = reverse_sequence(inputs=inputs,
                                      seq_lengths=sequence_length,
                                      time_major=time_major)

    tmp, output_state_bw = dynamic_rnn(cell=cell_bw,
                                       inputs=inputs_reverse,
                                       sequence_length=sequence_length,
                                       initial_state=initial_state_bw,
                                       time_major=time_major)
    output_bw = reverse_sequence(inputs=tmp,
                                 seq_lengths=sequence_length,
                                 time_major=time_major)

    outputs = (output_fw, output_bw)
    output_states = (output_state_fw, output_state_bw)

    return outputs, output_states


def dynamic_rnn(
        cell: RNNCellBase[State],
        inputs: torch.Tensor,
        sequence_length: Optional[Union[torch.LongTensor, List[int]]] = None,
        initial_state: Optional[State] = None,
        time_major: bool = False) -> Tuple[torch.Tensor, State]:
    r"""Creates a recurrent neural network specified by RNNCell ``cell``.

    Performs fully dynamic unrolling of ``inputs``.

    Args:
        cell: An instance of RNNCell.
        inputs: The RNN inputs.
            If ``time_major == False`` (default), this must be a ``Tensor``
            of shape: ``[batch_size, max_time, ...]``, or a nested
            tuple of such elements.
            If ``time_major == True``, this must be a ``Tensor`` of shape:
            ``[max_time, batch_size, ...]``, or a nested tuple of such
            elements.
            This may also be a (possibly nested) tuple of Tensors satisfying
            this property.  The first two dimensions must match across all the
            inputs, but otherwise the ranks and other shape components
            may differ. In this case, input to ``cell`` at each time-step
            will replicate the structure of these tuples, except for the
            time dimension (from which the time is taken).
            The input to ``cell`` at each time step will be a
            ``Tensor`` or (possibly nested) tuple of Tensors each with
            dimensions ``[batch_size, ...]``.
        sequence_length: (optional) An int32/int64 tensor sized
            ``[batch_size]``. Used to copy-through state and
            zero-out outputs when past a batch element's sequence length.
            So it's more for performance than correctness.
        initial_state: (optional) An initial state for the RNN.
            If ``cell.state_size`` is an integer, this must be
            a ``Tensor`` of appropriate type and shape
            ``[batch_size, cell.state_size]``. If ``cell.state_size`` is
            a tuple, this should be a tuple of tensors having shapes
            ``[batch_size, s]`` for ``s`` in ``cell.state_size``.
        time_major: The shape format of the ``inputs`` and ``outputs`` Tensors.
            If true, these ``Tensors`` must be shaped
            ``[max_time, batch_size, depth]``. If false, these ``Tensors``
            must be shaped ``[batch_size, max_time, depth]``.
            Using ``time_major = True`` is a bit more efficient because
            it avoids transposes at the beginning and end of the
            RNN calculation. However, most TensorFlow data is batch-major,
            so by default this function accepts input and emits output in
            batch-major form.

    Returns:
        A pair (outputs, state) where:

        outputs: The RNN output ``Tensor``.

            If time_major == False (default), this will be a ``Tensor`` shaped:
            ``[batch_size, max_time, cell.output_size]``.

            If time_major == True, this will be a ``Tensor`` shaped:
            ``[max_time, batch_size, cell.output_size]``.

            Note, if ``cell.output_size`` is a (possibly nested) tuple of
            integers or ``torch.Size`` objects, then ``outputs``
            will be a tuple having the same structure as ``cell.output_size``,
            containing Tensors having shapes corresponding to the shape
            data in ``cell.output_size``.

        state: The final state.  If ``cell.state_size`` is an int, this
            will be shaped ``[batch_size, cell.state_size]``.  If it is a
            ``torch.Size``, this will be shaped
            ``[batch_size] + cell.state_size``.
            If it is a (possibly nested) tuple of ints or ``torch.Size``,
            this will be a tuple having the corresponding shapes.
            If cells are ``LSTMCells``, ``state`` will be a tuple containing
            a ``LSTMStateTuple`` for each cell.

    Raises:
        TypeError: If ``cell`` is not an instance of RNNCell.
        ValueError: If inputs is None or an empty list.
    """
    # By default, time_major==False and inputs are batch-major: shaped
    #   [batch, time, depth]
    # For internal calculations, we transpose to [time, batch, depth]
    if not time_major:
        # (B,T,D) => (T,B,D)
        inputs = inputs.permute(1, 0, 2)

    time_steps = inputs.shape[0]
    batch_size = inputs.shape[1]

    if sequence_length is not None:
        if not isinstance(sequence_length, torch.Tensor):
            sequence_length = torch.tensor(sequence_length,
                                           dtype=torch.int32,
                                           device=inputs.device)

        if sequence_length.dim() != 1:
            raise ValueError(
                "sequence_length must be a vector of length batch_size, "
                "but saw shape: %s" % sequence_length.shape)
        if sequence_length.shape != torch.Size([batch_size]):
            raise ValueError("Expected shape for Tensor sequence_length is %s"
                             % batch_size, " but saw shape: %s"
                             % sequence_length.shape)
    else:
        sequence_length = torch.tensor([time_steps] * batch_size,
                                       dtype=torch.int32,
                                       device=inputs.device)

    if initial_state is not None:
        state = initial_state
    else:
        state = cell.zero_state(batch_size=batch_size)

    (outputs, final_state) = _dynamic_rnn_loop(
        cell, inputs, state, sequence_length=sequence_length)

    # Outputs of _dynamic_rnn_loop are always shaped [time, batch, depth].
    # If we are performing batch-major calculations, transpose output back
    # to shape [batch, time, depth]
    if not time_major:
        # (T,B,D) => (B,T,D)
        outputs = outputs.permute(1, 0, 2)

    return outputs, final_state


def _dynamic_rnn_loop(cell: RNNCellBase[State],
                      inputs: torch.Tensor,
                      initial_state: State,
                      sequence_length: torch.LongTensor) \
        -> Tuple[torch.Tensor, State]:
    r"""Internal implementation of Dynamic RNN.

    Args:
        cell: An instance of RNNCell.
        inputs: A ``Tensor`` of shape ``[time, batch_size, input_size]``,
            or a nested tuple of such elements.
        initial_state: A ``Tensor`` of shape ``[batch_size, state_size]``,
            or if ``cell.state_size`` is a tuple, then this should be a tuple
            of tensors having shapes ``[batch_size, s]`` for ``s`` in
            ``cell.state_size``.
        sequence_length: (optional) An ``int32`` ``Tensor``
            of shape ``[batch_size]``.

    Returns:
        Tuple ``(final_outputs, final_state)``.
        final_outputs:
            A ``Tensor`` of shape ``[time, batch_size, cell.output_size]``. If
            ``cell.output_size`` is a (possibly nested) tuple of ints or
            ``torch.Size`` objects, then this returns a
            (possibly nested) tuple of Tensors matching the corresponding
            shapes.
        final_state:
            A ``Tensor``, or possibly nested tuple of Tensors, matching
            in length and shapes to ``initial_state``.
    """
    state = initial_state
    time_steps = inputs.shape[0]
    all_outputs = []

    all_state = map_structure(lambda _: no_map(list), state)

    for i in range(time_steps):
        output, state = cell(inputs[i], state)
        all_outputs.append(output)
        map_structure_zip(lambda xs, x: xs.append(x), (all_state, state))
    # TODO: Do not compute everything regardless of sequence_length

    final_outputs = torch.stack(all_outputs, dim=0)
    final_outputs = mask_sequences(final_outputs,
                                   sequence_length=sequence_length,
                                   time_major=True)

    final_state = map_structure(lambda _: no_map(list), state)
    # pylint: disable=cell-var-from-loop
    # Our use case is fine because the function is called immediately and
    # exclusively in the current iteration of the loop.
    for batch_idx, time_idx in enumerate(sequence_length.tolist()):
        if time_idx > 0:
            map_structure_zip(
                lambda xs, x: xs.append(x[time_idx - 1][batch_idx]),
                (final_state, all_state))
        else:
            map_structure_zip(
                lambda xs, x: xs.append(x[batch_idx]),
                (final_state, initial_state))
    # pylint: enable=cell-var-from-loop

    final_state = map_structure(
            lambda x: torch.stack(x, dim=0), final_state)

    return final_outputs, final_state



# Copyright 2019 The Texar Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Unit tests for XLNet utils.
"""

import os
import unittest

from texar.torch.modules.pretrained.xlnet import *
from texar.torch.utils.test import pretrained_test


class XLNetUtilsTest(unittest.TestCase):
    r"""Tests XLNet utils.
    """

    @pretrained_test
    def test_load_pretrained_xlnet_AND_transform_xlnet_to_texar_config(self):

        pretrained_model_dir = PretrainedXLNetMixin.download_checkpoint(
            pretrained_model_name="xlnet-base-cased")

        info = list(os.walk(pretrained_model_dir))
        _, _, files = info[0]
        self.assertIn('spiece.model', files)
        self.assertIn('xlnet_model.ckpt.meta', files)
        self.assertIn('xlnet_model.ckpt.data-00000-of-00001', files)
        self.assertIn('xlnet_model.ckpt.index', files)
        self.assertIn('xlnet_config.json', files)

        model_config = PretrainedXLNetMixin._transform_config(
            pretrained_model_name="xlnet-base-cased",
            cache_dir=pretrained_model_dir)

        exp_config = {'head_dim': 64,
                      'ffn_inner_dim': 3072,
                      'hidden_dim': 768,
                      'activation': 'gelu',
                      'num_heads': 12,
                      'num_layers': 12,
                      'vocab_size': 32000,
                      'untie_r': True}

        self.assertDictEqual(model_config, exp_config)


if __name__ == "__main__":
    unittest.main()



# Copyright 2019 The Texar Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Various data classes that define data reading, parsing, batching, and other
preprocessing operations.
"""
from typing import (List, Optional, Union)

from distutils.util import strtobool
import numpy as np
import torch

from texar.torch.data.data.data_base import DatasetBase, DataSource
from texar.torch.data.data.dataset_utils import Batch
from texar.torch.data.data.text_data_base import TextLineDataSource
from texar.torch.hyperparams import HParams
from texar.torch.utils.dtypes import get_numpy_dtype, \
    get_supported_scalar_types, torch_bool


__all__ = [
    "_default_scalar_dataset_hparams",
    "ScalarData"
]


def _default_scalar_dataset_hparams():
    """Returns hyperparameters of a scalar dataset with default values.

    See :meth:`texar.torch.data.ScalarData.default_hparams` for details.
    """
    return {
        "files": [],
        "compression_type": None,
        "data_type": "int",
        "data_name": "data",
        "other_transformations": [],
        "@no_typecheck": ["files"]
    }


class ScalarData(DatasetBase[List[str], Union[int, float]]):
    r"""Scalar data where each line of the files is a scalar (int or float),
    e.g., a data label.

    Args:
        hparams (dict): Hyperparameters. See :meth:`default_hparams` for the
            defaults.
        device: The device of the produced batches. For GPU training, set to
            current CUDA device.

    The processor reads and processes raw data and results in a dataset
    whose element is a python `dict` including one field. The field name is
    specified in :attr:`hparams["dataset"]["data_name"]`. If not specified,
    the default name is `"data"`. The field name can be accessed through
    :attr:`data_name`.

    This field is a Tensor of shape `[batch_size]` containing a batch of
    scalars, of either int or float type as specified in :attr:`hparams`.

    Example:

        .. code-block:: python

            hparams={
                'dataset': { 'files': 'data.txt', 'data_name': 'label' },
                'batch_size': 2
            }
            data = ScalarData(hparams)
            iterator = DataIterator(data)
            for batch in iterator:
                # batch contains the following
                # batch == {
                #     'label': [2, 9]
                # }
    """

    def __init__(self, hparams, device: Optional[torch.device] = None,
                 data_source: Optional[DataSource] = None):
        self._hparams = HParams(hparams, self.default_hparams())
        self._other_transforms = self._hparams.dataset.other_transformations
        data_type = self._hparams.dataset["data_type"]
        if data_type not in get_supported_scalar_types():
            raise ValueError(f"Unsupported data type '{data_type}'")

        # In Pytorch versions < 1.1.0, "torch.uint8" is treated as "bool" type
        # hence we set self.data_type = np.uint8 here
        if data_type == "bool":
            self._data_type = get_numpy_dtype(str(torch_bool))
        else:
            self._data_type = get_numpy_dtype(data_type)

        if data_source is None:
            data_source = TextLineDataSource(
                self._hparams.dataset.files,
                compression_type=self._hparams.dataset.compression_type)
        super().__init__(data_source, hparams, device=device)

    @staticmethod
    def default_hparams():
        r"""Returns a dictionary of default hyperparameters.

        .. code-block:: python

            {
                # (1) Hyperparams specific to scalar dataset
                "dataset": {
                    "files": [],
                    "compression_type": None,
                    "data_type": "int",
                    "other_transformations": [],
                    "data_name": "data",
                }
                # (2) General hyperparams
                "num_epochs": 1,
                "batch_size": 64,
                "allow_smaller_final_batch": True,
                "shuffle": True,
                "shuffle_buffer_size": None,
                "shard_and_shuffle": False,
                "num_parallel_calls": 1,
                "prefetch_buffer_size": 0,
                "max_dataset_size": -1,
                "seed": None,
                "name": "scalar_data",
            }

        Here:

        1. For the hyperparameters in the :attr:`"dataset"` field:

            `"files"`: str or list
                A (list of) file path(s).

                Each line contains a single scalar number.

            `"compression_type"`: str, optional
                One of "" (no compression), "ZLIB", or "GZIP".

            `"data_type"`: str
                The scalar type. Types defined in
                :meth:`~texar.torch.utils.dtypes.get_supported_scalar_types` are
                supported.

            `"other_transformations"`: list
                A list of transformation functions or function names/paths to
                further transform each single data instance.

                (More documentations to be added.)

            `"data_name"`: str
                Name of the dataset.

        2. For the **general** hyperparameters, see
           :meth:`texar.torch.data.DatasetBase.default_hparams` for details.

        """
        hparams = DatasetBase.default_hparams()
        hparams["name"] = "scalar_data"
        hparams.update({
            "dataset": _default_scalar_dataset_hparams()
        })
        return hparams

    def process(self, raw_example: List[str]) -> Union[bool, int, float]:
        assert len(raw_example) == 1

        example_: Union[int, str]
        if self._data_type == np.bool_:
            example_ = strtobool(raw_example[0])

        else:
            example_ = raw_example[0]

        example = self._data_type(example_)

        for transform in self._other_transforms:
            example = transform(example)
        return example

    def collate(self, examples: List[Union[bool, int, float]]) -> Batch:
        # convert the list of strings into appropriate tensors here
        examples_np = np.array(examples, dtype=self._data_type)
        collated_examples = torch.from_numpy(examples_np)
        return Batch(len(examples),
                     batch={self.data_name: collated_examples})

    def list_items(self):
        r"""Returns the list of item names that the data can produce.

        Returns:
            A list of strings.
        """
        return [self.hparams.dataset["data_name"]]

    @property
    def data_name(self):
        r"""The name of the data tensor, "data" by default if not specified in
        :attr:`hparams`.
        """
        return self.hparams.dataset["data_name"]



# Copyright 2019 The Texar Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Executor metrics for classification tasks.
"""

from abc import ABC
from typing import Dict, List, Optional, Sequence, Tuple, TypeVar

import numpy as np

from texar.torch.run.metric.base_metric import StreamingMetric

__all__ = [
    "Accuracy",
    "ConfusionMatrix",
    "Precision",
    "Recall",
    "F1",
]

Input = TypeVar('Input')
Value = TypeVar('Value')


class Accuracy(StreamingMetric[Input, float]):
    r"""The accuracy metric for evaluation classification tasks. Accuracy is
    defined as the ratio of correct (exactly matching) predictions out of all
    predictions.

    Accuracy is a :class:`~texar.torch.run.metric.StreamingMetric`, requires
    both predicted values and labels. Accuracy values are :class:`float`
    numbers between 0 and 1, with higher values being better.

    Keyword Args:
        pred_name (str): Name of the predicted value. This will be used as the
            key to the dictionary returned by the model.
        label_name (str): Name of the label. This will be used as the key to the
            batch object returned by the dataset. Defaults to ``"label"``.
    """
    correct: int

    def reset(self) -> None:
        super().reset()
        self.correct = 0

    def add(self, predicted: Sequence[Input], labels: Sequence[Input]) -> None:
        super().add(predicted, labels)
        self.correct += sum(int(a == b) for a, b in zip(predicted, labels))

    def value(self) -> float:
        if self.count == 0:
            return 0.0
        return self.correct / self.count


class _ConfusionMatrix(StreamingMetric[Input, Value], ABC):
    count: int
    matrix: Optional[np.ndarray]  # matrix[pred][label]
    pred_count: List[int]
    label_count: List[int]
    _class_id: Dict[Input, int]

    def reset(self) -> None:
        super().reset()
        self.matrix = None
        self.pred_count = []
        self.label_count = []
        self._class_id = {}

    def _convert_ids(self, classes: Sequence[Input]) -> List[int]:
        ids = []
        cnt = 0
        for klass in classes:
            if klass not in self._class_id:
                self._class_id[klass] = len(self._class_id)
                cnt += 1
            ids.append(self._class_id[klass])
        if self.matrix is None:
            self.matrix = np.zeros((cnt, cnt), dtype=np.int)
        else:
            self.matrix = np.pad(self.matrix, [(0, cnt), (0, cnt)],
                                 "constant", constant_values=0)
        self.pred_count.extend([0] * cnt)
        self.label_count.extend([0] * cnt)
        return ids

    def add(self, predicted: Sequence[Input], labels: Sequence[Input]) -> None:
        super().add(predicted, labels)
        predicted = self._convert_ids(predicted)
        labels = self._convert_ids(labels)
        assert self.matrix is not None
        for pred, label in zip(predicted, labels):
            self.matrix[pred, label] += 1
            self.pred_count[pred] += 1
            self.label_count[label] += 1


class ConfusionMatrix(_ConfusionMatrix[Input, Optional[np.ndarray]]):
    r"""The confusion matrix is an evaluation metric for classification tasks.

    Confusion matrix is a :class:`~texar.torch.run.metric.StreamingMetric`,
    requires both predicted values and labels. Confusion matrix values are NumPy
    arrays, with no clear definition of "better". Comparison of two confusion
    matrices are not meaningful.

    The value indexed at ``(i, j)`` of the confusion matrix is the number of
    data points whose predicted label is `i` and whose ground truth label is
    `j`. Labels are internally mapped to indices.

    Keyword Args:
        pred_name (str): Name of the predicted value. This will be used as the
            key to the dictionary returned by the model.
        label_name (str): Name of the label. This will be used as the key to the
            batch object returned by the dataset. Defaults to ``"label"``.
    """

    def value(self) -> Optional[np.ndarray]:
        return self.matrix

    @property
    def class_id(self):
        r"""Mapping of predicted values and labels to indices within the matrix.
        """
        return self._class_id

    def better(self, cur: Value, prev: Value) -> Optional[bool]:
        # Always return `None` to indicate values are uncomparable.
        return None


class _MicroMacro(_ConfusionMatrix[Input, float], ABC):
    _valid_modes = ['binary', 'micro', 'macro', 'weighted']

    def __init__(self, mode: str = 'binary', pos_label: Optional[Input] = None,
                 *, pred_name: str, label_name: str = "label"):
        super().__init__(pred_name=pred_name, label_name=label_name)
        self.mode = mode
        if self.mode not in self._valid_modes:
            raise ValueError(f"Invalid mode {mode}. "
                             f"Supported modes are: {self._valid_modes}")
        if self.mode == 'binary' and pos_label is None:
            raise ValueError("`pos_label` must not be none when `mode` is "
                             "set to 'binary'")
        if pos_label is not None:
            self.pos_label = pos_label

    def _safe_divide(self, numerator: np.ndarray, denominator: np.ndarray) \
            -> np.ndarray:
        # Credit: sklearn.metrics.classification._prf_divide
        if numerator.size == 1:
            if denominator == 0.0:
                return np.array(0.0)
            return numerator / denominator

        mask = denominator == 0.0
        denominator = denominator.copy()
        denominator[mask] = 1.0
        value = numerator / denominator
        return value

    def _convert_value(self, value: np.ndarray) -> np.ndarray:
        if self.mode == 'binary':
            label = self._class_id.get(self.pos_label, None)
            if label is None:
                return np.array(0)
            return value[label]
        if self.mode == 'micro':
            return value.sum()
        return value

    def _true_positive(self) -> np.ndarray:
        assert self.matrix is not None
        value = self.matrix.diagonal()
        return self._convert_value(value)

    def _true_negative(self) -> np.ndarray:
        assert self.matrix is not None
        value = (self.count
                 - np.asarray(self.pred_count)
                 - np.asarray(self.label_count)
                 + self.matrix.diagonal())
        return self._convert_value(value)

    def _false_positive(self) -> np.ndarray:
        assert self.matrix is not None
        value = np.asarray(self.pred_count) - self.matrix.diagonal()
        return self._convert_value(value)

    def _false_negative(self) -> np.ndarray:
        assert self.matrix is not None
        value = np.asarray(self.label_count) - self.matrix.diagonal()
        return self._convert_value(value)

    def _value(self) -> Tuple[np.ndarray, np.ndarray]:
        r"""Return the numerator and denominator of the metric value.
        """
        raise NotImplementedError

    def value(self) -> float:
        if self.count == 0:
            return 0.0
        numerator, denominator = self._value()
        value = self._safe_divide(numerator, denominator)
        if self.mode == 'macro':
            value = value.sum() / len(self._class_id)
        elif self.mode == 'weighted':
            value = (value * np.asarray(self.label_count)).sum() / self.count
        return value.item()


class Precision(_MicroMacro[Input]):
    r"""The precision metric for evaluation classification tasks. Precision is
    defined as the ratio of ``tp / (tp + fp)``, where ``tp`` is the number of
    true positives and ``fp`` is the number of false positives.

    Precision is a :class:`~texar.torch.run.metric.StreamingMetric`, requires
    both predicted values and labels. Precision values are :class:`float`
    numbers between 0 and 1, with higher values being better.

    Args:
        mode (str): The mode for computing averages across multiple labels.
            Defaults to ``"binary"``. Available options include:

            - ``"binary"``: Only report results for the class specified by
              :attr:`pos_label`. This is only meaningful for binary
              classification tasks.
            - ``"micro"``: Return the precision value computed using globally
              counted true positives and false positives.
            - ``"macro"``: Return the unweighted average of precision values for
              each label.
            - ``"weighted"``: Return the average of precision values for each
              label, weighted by the number of true instances for each label.
        pos_label (str, optional): The label for the positive class. Only used
            if :attr:`mode` is set to ``"binary"``.

    Keyword Args:
        pred_name (str): Name of the predicted value. This will be used as the
            key to the dictionary returned by the model.
        label_name (str): Name of the label. This will be used as the key to the
            batch object returned by the dataset. Defaults to ``"label"``.
    """

    def _value(self) -> Tuple[np.ndarray, np.ndarray]:
        return (self._true_positive(),
                (self._true_positive() + self._false_positive()))


class Recall(_MicroMacro[Input]):
    r"""The recall metric for evaluation classification tasks. Precision is
    defined as the ratio of ``tp / (tp + fn)``, where ``tp`` is the number of
    true positives and ``fn`` is the number of false negatives.

    Recall is a :class:`~texar.torch.run.metric.StreamingMetric`, requires both
    predicted values and labels. Recall values are :class:`float` numbers
    between 0 and 1, with higher values being better.

    Args:
        mode (str): The mode for computing averages across multiple labels.
            Defaults to ``"binary"``. Available options include:

            - ``"binary"``: Only report results for the class specified by
              :attr:`pos_label`. This is only meaningful for binary
              classification tasks.
            - ``"micro"``: Return the recall value computed using globally
              counted true positives and false negatives.
            - ``"macro"``: Return the unweighted average of recall values for
              each label.
            - ``"weighted"``: Return the average of recall values for each
              label, weighted by the number of true instances for each label.
        pos_label (str, optional): The label for the positive class. Only used
            if :attr:`mode` is set to ``"binary"``.

    Keyword Args:
        pred_name (str): Name of the predicted value. This will be used as the
            key to the dictionary returned by the model.
        label_name (str): Name of the label. This will be used as the key to the
            batch object returned by the dataset. Defaults to ``"label"``.
    """

    def _value(self) -> Tuple[np.ndarray, np.ndarray]:
        return (self._true_positive(),
                (self._true_positive() + self._false_negative()))


class F1(Precision[Input], Recall[Input]):
    r"""The F1 metric for evaluation classification tasks. F1 is defined as the
    harmonic mean of precision and recall.

    F1 is a :class:`~texar.torch.run.metric.StreamingMetric`, requires both
    predicted values and labels. F1 values are :class:`float` numbers between 0
    and 1, with higher values being better.

    Args:
        mode (str): The mode for computing averages across multiple labels.
            Defaults to ``"binary"``. Available options include:

            - ``"binary"``: Only report results for the class specified by
              :attr:`pos_label`. This is only meaningful for binary
              classification tasks.
            - ``"micro"``: Return the F1 value computed using globally counted
              true positives, false positives, and false negatives.
            - ``"macro"``: Return the unweighted average of F1 values for each
              label.
            - ``"weighted"``: Return the average of F1 values for each label,
              weighted by the number of true instances for each label.
        pos_label (str, optional): The label for the positive class. Only used
            if :attr:`mode` is set to ``"binary"``.

    Keyword Args:
        pred_name (str): Name of the predicted value. This will be used as the
            key to the dictionary returned by the model.
        label_name (str): Name of the label. This will be used as the key to the
            batch object returned by the dataset. Defaults to ``"label"``.
    """

    def _value(self) -> Tuple[np.ndarray, np.ndarray]:
        # pylint: disable=protected-access
        precision = self._safe_divide(*Precision._value(self))
        recall = self._safe_divide(*Recall._value(self))
        # pylint: enable=protected-access
        return (2 * precision * recall,
                precision + recall)



# Copyright 2019 The Texar Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Python implementation of BLEU adapted from:
    `https://github.com/tensorflow/models/blob/master/official/transformer/compute_bleu.py`
"""

from typing import Callable, Counter, List, Tuple

import re
import sys
import unicodedata
import collections
import math
import numpy as np

from texar.torch.evals.bleu import corpus_bleu
from texar.torch.evals.bleu_moses import corpus_bleu_moses
from texar.torch.utils.types import MaybeList

__all__ = [
    "corpus_bleu_transformer",
    "bleu_transformer_tokenize",
    "file_bleu",
]


def _get_ngrams(segment: List[str],
                max_order: int) -> Counter[Tuple[str, ...]]:
    r"""Extracts all n-grams up to a given maximum order from an
    input segment.

    Args:
        segment: text segment from which n-grams will be extracted.
        max_order: maximum length in tokens of the n-grams returned
            by this methods.

    Returns:
        The Counter containing all n-grams upto :attr:`max_order`
        in segment with a count of how many times each n-gram occurred.
    """
    ngram_counts: collections.Counter = collections.Counter()
    for order in range(1, max_order + 1):
        for i in range(0, len(segment) - order + 1):
            ngram = tuple(segment[i:i + order])
            ngram_counts[ngram] += 1
    return ngram_counts


def corpus_bleu_transformer(reference_corpus: List[List[str]],
                            translation_corpus: List[List[str]],
                            max_order: int = 4,
                            use_bp: bool = True) -> float:
    r"""Computes BLEU score of translated segments against references.

    This BLEU has been used in evaluating Transformer (Vaswani et al.)
    "Attention is all you need" for machine translation. The resulting BLEU
    score are usually a bit higher than that in
    `texar.torch.evals.corpus_bleu` and `texar.torch.evals.corpus_bleu_moses`.

    Args:
        reference_corpus: list of references for each translation. Each
            reference should be tokenized into a list of tokens.
        translation_corpus: list of translations to score. Each translation
            should be tokenized into a list of tokens.
        max_order: Maximum n-gram order to use when computing BLEU score.
        use_bp: boolean, whether to apply brevity penalty.

    Returns:
        BLEU score.
    """
    reference_length = 0
    translation_length = 0
    bp = 1.0
    geo_mean = 0.0

    matches_by_order = [0] * max_order
    possible_matches_by_order = [0] * max_order

    for (references, translations) in zip(reference_corpus, translation_corpus):
        reference_length += len(references)
        translation_length += len(translations)
        ref_ngram_counts = _get_ngrams(references, max_order)
        translation_ngram_counts = _get_ngrams(translations, max_order)

        overlap = dict(
            (ngram, min(count, translation_ngram_counts[ngram]))
            for ngram, count in ref_ngram_counts.items()
        )

        for ngram in overlap:
            matches_by_order[len(ngram) - 1] += overlap[ngram]
        for ngram in translation_ngram_counts:
            possible_matches_by_order[len(ngram) - 1] += \
                translation_ngram_counts[ngram]

    precisions = [0.0] * max_order
    smooth = 1.0

    for i in range(max_order):
        if possible_matches_by_order[i] > 0:
            precisions[i] = matches_by_order[i] / possible_matches_by_order[i]
            if matches_by_order[i] > 0:
                precisions[i] = (matches_by_order[i] /
                                 possible_matches_by_order[i])
            else:
                smooth *= 2
                precisions[i] = 1.0 / (smooth * possible_matches_by_order[i])
        else:
            precisions[i] = 0.0

    if max(precisions) > 0:
        p_log_sum = sum(math.log(p) for p in precisions if p)
        geo_mean = math.exp(p_log_sum / max_order)

    if use_bp:
        ratio = translation_length / reference_length
        if ratio <= 0:
            bp = 0
        elif ratio < 1.0:
            bp = math.exp(1 - 1.0 / ratio)
        else:
            bp = 1.0

    bleu = geo_mean * bp

    return np.float32(bleu) * 100


class UnicodeRegex:
    """Ad-hoc hack to recognize all punctuation and symbols."""

    def __init__(self):
        punctuation = self.property_chars("P")
        self.nondigit_punct_re = re.compile(r"([^\d])([" + punctuation + r"])")
        self.punct_nondigit_re = re.compile(r"([" + punctuation + r"])([^\d])")
        self.symbol_re = re.compile("([" + self.property_chars("S") + "])")

    @staticmethod
    def property_chars(prefix):
        return "".join(
            chr(x)
            for x in range(sys.maxunicode)
            if unicodedata.category(chr(x)).startswith(prefix)
        )


uregex = UnicodeRegex()


def bleu_transformer_tokenize(string: str) -> List[str]:
    r"""Tokenize a string following the official BLEU implementation.

    The BLEU scores from `multi-bleu.perl` depend on your `tokenizer`, which is
    unlikely to be reproducible from your experiment or consistent across
    different users. This function provides a standard tokenization following
    `mteval-v14.pl`.

    See
    `https://github.com/moses-smt/mosesdecoder/blob/master/scripts/generic/mteval-v14.pl#L954-L983`.
    In our case, the input string is expected to be just one line
    and no HTML entities de-escaping is needed.
    So we just tokenize on punctuation and symbols,
    except when a punctuation is preceded and followed by a digit
    (e.g. a comma/dot as a thousand/decimal separator).

    Note that a number (e.g. a year) followed by a dot at the end of sentence
    is NOT tokenized,
    i.e. the dot stays with the number because `s/(\p{P})(\P{N})/ $1 $2/g`
    does not match this case (unless we add a space after each sentence).
    However, this error is already in the original `mteval-v14.pl`
    and we want to be consistent with it.

    Args:
        string: the input string

    Returns:
        a list of tokens
    """
    string = uregex.nondigit_punct_re.sub(r"\1 \2 ", string)
    string = uregex.punct_nondigit_re.sub(r" \1 \2", string)
    string = uregex.symbol_re.sub(r" \1 ", string)
    return string.split()


def file_bleu(ref_filename: str,
              hyp_filename: str,
              bleu_version: str = "corpus_bleu_transformer",
              case_sensitive: bool = False) -> float:
    r"""Compute BLEU for two files (reference and hypothesis translation).

    Args:
        ref_filename: Reference file path.
        hyp_filename: Hypothesis file path.
        bleu_version: A str with the name of a BLEU computing method selected
            in the list of: `corpus_bleu`, `corpus_bleu_moses`,
            `corpus_bleu_transformer`.
        case_sensitive: If `False`, lowercase reference and hypothesis
            tokens.

    Returns:
        BLEU score.
    """
    with open(ref_filename, encoding="utf-8") as f:
        ref_lines = f.read().splitlines()

    with open(hyp_filename, encoding="utf-8") as f:
        hyp_lines = f.read().splitlines()

    if len(ref_lines) != len(hyp_lines):
        raise ValueError(
            "Reference and translation files have different number of lines")

    if not case_sensitive:
        ref_lines = [x.lower() for x in ref_lines]
        hyp_lines = [x.lower() for x in hyp_lines]

    ref_tokens: List[MaybeList[List[str]]]
    if bleu_version == "corpus_bleu_transformer":
        ref_tokens = [bleu_transformer_tokenize(x) for x in ref_lines]
    else:
        ref_tokens = [[bleu_transformer_tokenize(x)] for x in ref_lines]
    hyp_tokens = [bleu_transformer_tokenize(x) for x in hyp_lines]

    bleu_dict = {
        "corpus_bleu": corpus_bleu,
        "corpus_bleu_moses": corpus_bleu_moses,
        "corpus_bleu_transformer": corpus_bleu_transformer
    }
    fn: Callable[[List[MaybeList[List[str]]], List[List[str]]], float]
    fn = bleu_dict[bleu_version]  # type: ignore

    return fn(ref_tokens, hyp_tokens)



# Copyright 2019 The Texar Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Various neural network layers
"""

import copy
import functools
from typing import Any, Callable, Dict, List, Optional, Type, Union

import torch
from torch import nn

from texar.torch.core import cell_wrappers as wrappers
from texar.torch.core.regularizers import L1L2, Regularizer
from texar.torch.hyperparams import HParams
from texar.torch.utils import utils
from texar.torch.utils.dtypes import is_str

__all__ = [
    'default_rnn_cell_hparams',
    'get_rnn_cell',
    'identity',
    'default_regularizer_hparams',
    'get_initializer',
    'get_regularizer',
    'get_activation_fn',
    'get_layer',
    'MaxReducePool1d',
    'AvgReducePool1d',
    'get_pooling_layer_hparams',
    'MergeLayer',
    'Flatten',
    'Identity',
]


def default_rnn_cell_hparams():
    r"""Returns a `dict` of RNN cell hyperparameters and their default values.

    .. code-block:: python

        {
            "type": "LSTMCell",
            "input_size": 256,
            "kwargs": {
                "hidden_size": 256
            },
            "num_layers": 1,
            "dropout": {
                "input_keep_prob": 1.0,
                "output_keep_prob": 1.0,
                "state_keep_prob": 1.0,
                "variational_recurrent": False,
            },
            "residual": False,
            "highway": False,
        }

    Here:

    `"type"`: str or cell class or cell instance
        The RNN cell type. This can be

        - The string name or full module path of a cell class. If class name is
          provided, the class must be in module :mod:`torch.nn.modules.rnn`,
          :mod:`texar.torch.core.cell_wrappers`, or :mod:`texar.torch.custom`.
        - A cell class.
        - An instance of a cell class. This is not valid if `"num_layers"` > 1.

        For example

        .. code-block:: python

            "type": "LSTMCell"  # class name
            "type": "torch.nn.GRUCell"  # module path
            "type": "my_module.MyCell"  # module path
            "type": torch.nn.GRUCell  # class
            "type": LSTMCell(hidden_size=100)  # cell instance
            "type": MyCell(...)  # cell instance

    `"kwargs"`: dict
        Keyword arguments for the constructor of the cell class.
        A cell is created by :python:`cell_class(**kwargs)`, where
        `cell_class` is specified in "type" above.

        Ignored if "type" is a cell instance.

        .. note::
            It is unnecessary to specify `"input_size"` within `"kwargs"`.
            This value will be automatically filled based on layer index.

        .. note::
            Although PyTorch uses `"hidden_size"` to denote the hidden layer
            size, we follow TensorFlow conventions and use `"num_units"`.

    `"num_layers"`: int
        Number of cell layers. Each layer is a cell created as above, with
        the same hyperparameters specified in `"kwargs"`.

    `"dropout"`: dict
        Dropout applied to the cell in **each** layer. See
        :class:`~texar.torch.core.cell_wrappers.DropoutWrapper` for details of
        the hyperparameters. If all `"\*_keep_prob"` = 1, no dropout is applied.

        Specifically, if `"variational_recurrent"` = `True`,
        the same dropout mask is applied across all time steps per batch.

    `"residual"`: bool
        If `True`, apply residual connection on the inputs and
        outputs of cell in **each** layer except the first layer. Ignored
        if `"num_layers"` = 1.

    `"highway"`: bool
        If True, apply highway connection on the inputs and
        outputs of cell in each layer except the first layer. Ignored if
        `"num_layers"` = 1.
    """
    return {
        'type': 'LSTMCell',
        'kwargs': {
            'num_units': 256,
        },
        'num_layers': 1,
        'dropout': {
            'input_keep_prob': 1.0,
            'output_keep_prob': 1.0,
            'state_keep_prob': 1.0,
            'variational_recurrent': False,
        },
        'residual': False,
        'highway': False,
        '@no_typecheck': ['type']
    }


def default_regularizer_hparams():
    r"""Returns the hyperparameters and their default values of a variable
    regularizer:

    .. code-block:: python

        {
            "type": "L1L2",
            "kwargs": {
                "l1": 0.,
                "l2": 0.
            }
        }

    The default value corresponds to
    :class:`~texar.torch.core.regularizers.L1L2` and, with ``(l1=0, l2=0)``,
    disables regularization.
    """
    return {
        "type": "L1L2",
        "kwargs": {
            "l1": 0.,
            "l2": 0.
        }
    }


def get_rnn_cell(input_size, hparams=None):
    r"""Creates an RNN cell.

    See :func:`~texar.torch.core.default_rnn_cell_hparams` for all
    hyperparameters and default values.

    Args:
        input_size (int): Size of the input to the cell in the first layer.
        hparams (dict or HParams, optional): Cell hyperparameters. Missing
            hyperparameters are set to default values.

    Returns:
        A cell instance.

    Raises:
        ValueError: If ``hparams["num_layers"]``>1 and ``hparams["type"]`` is a
            class instance.
    """
    if hparams is None or isinstance(hparams, dict):
        hparams = HParams(hparams, default_rnn_cell_hparams())

    d_hp = hparams['dropout']
    variational_recurrent = d_hp['variational_recurrent']
    input_keep_prob = d_hp['input_keep_prob']
    output_keep_prob = d_hp['output_keep_prob']
    state_keep_prob = d_hp['state_keep_prob']

    cells = []
    num_layers = hparams['num_layers']
    cell_kwargs = hparams['kwargs'].todict()
    # rename 'num_units' to 'hidden_size' following PyTorch conventions
    cell_kwargs['hidden_size'] = cell_kwargs['num_units']
    del cell_kwargs['num_units']

    for layer_i in range(num_layers):
        # Create the basic cell
        cell_type = hparams["type"]
        if layer_i == 0:
            cell_kwargs['input_size'] = input_size
        else:
            cell_kwargs['input_size'] = cell_kwargs['hidden_size']
        if not isinstance(cell_type, str) and not isinstance(cell_type, type):
            if num_layers > 1:
                raise ValueError(
                    "If 'num_layers'>1, then 'type' must be a cell class or "
                    "its name/module path, rather than a cell instance.")
        cell_modules = ['texar.torch.core.cell_wrappers',  # prefer our wrappers
                        'torch.nn.modules.rnn', 'texar.torch.custom']
        cell = utils.check_or_get_instance(cell_type, cell_kwargs, cell_modules)
        if isinstance(cell, nn.RNNCellBase):
            cell = wrappers.wrap_builtin_cell(cell)

        # Optionally add dropout
        if (input_keep_prob < 1.0 or
                output_keep_prob < 1.0 or
                state_keep_prob < 1.0):
            # TODO: Would this result in non-final layer outputs being
            #       dropped twice?
            cell = wrappers.DropoutWrapper(
                cell=cell,
                input_keep_prob=input_keep_prob,
                output_keep_prob=output_keep_prob,
                state_keep_prob=state_keep_prob,
                variational_recurrent=variational_recurrent)

        # Optionally add residual and highway connections
        if layer_i > 0:
            if hparams['residual']:
                cell = wrappers.ResidualWrapper(cell)
            if hparams['highway']:
                cell = wrappers.HighwayWrapper(cell)

        cells.append(cell)

    if hparams['num_layers'] > 1:
        cell = wrappers.MultiRNNCell(cells)
    else:
        cell = cells[0]

    return cell


def identity(inputs: torch.Tensor):
    r"""Returns a tensor with the same content as the input tensor.

    Arguments:
        inputs: The input tensor.

    Returns:
        A tensor of the same shape, type, and content.
    """
    return inputs


def get_regularizer(hparams=None):
    r"""Returns a variable regularizer instance.

    See :func:`~texar.torch.core.default_regularizer_hparams` for all
    hyperparameters and default values.

    The "type" field can be a subclass
    of :class:`~texar.torch.core.regularizers.Regularizer`, its string name
    or module path, or a class instance.

    Args:
        hparams (dict or HParams, optional): Hyperparameters. Missing
            hyperparameters are set to default values.

    Returns:
        A :class:`~texar.torch.core.regularizers.Regularizer` instance.
        `None` if :attr:`hparams` is `None` or taking the default
        hyperparameter value.

    Raises:
        ValueError: The resulting regularizer is not an instance of
            :class:`~texar.torch.core.regularizers.Regularizer`.
    """

    if hparams is None:
        return None

    if isinstance(hparams, dict):
        hparams = HParams(hparams, default_regularizer_hparams())

    rgl = utils.check_or_get_instance(
        hparams.type, hparams.kwargs.todict(),
        ["texar.torch.core.regularizers", "texar.torch.custom"])

    if not isinstance(rgl, Regularizer):
        raise ValueError("The regularizer must be an instance of "
                         "texar.torch.core.regularizers.Regularizer.")

    if isinstance(rgl, L1L2) and rgl.l1 == 0. and rgl.l2 == 0.:
        return None

    return rgl


def get_initializer(hparams=None) \
        -> Optional[Callable[[torch.Tensor], torch.Tensor]]:
    r"""Returns an initializer instance.

    Args:
        hparams (dict or HParams, optional): Hyperparameters with the structure

            .. code-block:: python

                {
                    "type": "initializer_class_or_function",
                    "kwargs": {
                        # ...
                    }
                }

            The `"type"` field can be a function name or module path. If name is
            provided, it be must be from one the following modules:
            :torch_docs:`torch.nn.init <nn.html#torch-nn-init>` and
            :mod:`texar.torch.custom`.

            Besides, the `"type"` field can also be an initialization function
            called with :python:`initialization_fn(**kwargs)`. In this case
            `"type"` can be the function, or its name or module path. If no
            keyword argument is required, `"kwargs"` can be omitted.

    Returns:
        An initializer instance. `None` if :attr:`hparams` is `None`.
    """
    if hparams is None:
        return None

    kwargs = hparams.get('kwargs', {})
    if isinstance(kwargs, HParams):
        kwargs = kwargs.todict()
    modules = ['torch.nn.init', 'torch', 'texar.torch.custom']
    initializer_fn = utils.get_function(hparams['type'], modules)
    initializer = functools.partial(initializer_fn, **kwargs)

    return initializer


def get_activation_fn(fn_name: Optional[Union[str,
                                              Callable[[torch.Tensor],
                                                       torch.Tensor]]] = None,
                      kwargs: Union[HParams, Dict, None] = None) \
        -> Optional[Callable[[torch.Tensor], torch.Tensor]]:
    r"""Returns an activation function `fn` with the signature
    `output = fn(input)`.

    If the function specified by :attr:`fn_name` has more than one arguments
    without default values, then all these arguments except the input feature
    argument must be specified in :attr:`kwargs`. Arguments with default values
    can also be specified in :attr:`kwargs` to take values other than the
    defaults. In this case a partial function is returned with the above
    signature.

    Args:
        fn_name (str or callable): An activation function, or its name or
            module path. The function can be:

            - Built-in function defined in
              :torch_docs:`torch.nn.functional<nn.html#torch-nn-functional>`
            - User-defined activation functions in module
              :mod:`texar.torch.custom`.
            - External activation functions. Must provide the full module path,
              e.g., ``"my_module.my_activation_fn"``.

        kwargs (optional): A `dict` or instance of :class:`~texar.torch.HParams`
            containing the keyword arguments of the activation function.

    Returns:
        An activation function. `None` if :attr:`fn_name` is `None`.
    """
    if fn_name is None:
        return None

    fn_modules = ['torch', 'torch.nn.functional',
                  'texar.torch.custom', 'texar.torch.core.layers']
    activation_fn_ = utils.get_function(fn_name, fn_modules)
    activation_fn = activation_fn_

    # Make a partial function if necessary
    if kwargs is not None:
        if isinstance(kwargs, HParams):
            kwargs = kwargs.todict()

        def _partial_fn(features):
            return activation_fn_(features, **kwargs)

        activation_fn = _partial_fn

    return activation_fn


def get_layer(hparams: Union[HParams, Dict[str, Any]]) -> nn.Module:
    r"""Makes a layer instance.

    The layer must be an instance of :torch_nn:`Module`.

    Args:
        hparams (dict or HParams): Hyperparameters of the layer, with
            structure:

            .. code-block:: python

                {
                    "type": "LayerClass",
                    "kwargs": {
                        # Keyword arguments of the layer class
                        # ...
                    }
                }

            Here:

            `"type"`: str or layer class or layer instance
                The layer type. This can be

                - The string name or full module path of a layer class. If
                  the class name is provided, the class must be in module
                  :torch_nn:`Module`, :mod:`texar.torch.core`, or
                  :mod:`texar.torch.custom`.
                - A layer class.
                - An instance of a layer class.

                For example

                .. code-block:: python

                    "type": "Conv1D"                               # class name
                    "type": "texar.torch.core.MaxReducePooling1D"  # module path
                    "type": "my_module.MyLayer"                    # module path
                    "type": torch.nn.Module.Linear                 # class
                    "type": Conv1D(filters=10, kernel_size=2)  # cell instance
                    "type": MyLayer(...)                       # cell instance

            `"kwargs"`: dict
                A dictionary of keyword arguments for constructor of the
                layer class. Ignored if :attr:`"type"` is a layer instance.

                - Arguments named "activation" can be a callable, or a `str` of
                  the name or module path to the activation function.
                - Arguments named "\*_regularizer" and "\*_initializer" can be a
                  class instance, or a `dict` of hyperparameters of respective
                  regularizers and initializers. See
                - Arguments named "\*_constraint" can be a callable, or a `str`
                  of the name or full path to the constraint function.

    Returns:
        A layer instance. If ``hparams["type"]`` is a layer instance, returns it
        directly.

    Raises:
        ValueError: If :attr:`hparams` is `None`.
        ValueError: If the resulting layer is not an instance of
            :torch_nn:`Module`.
    """
    if hparams is None:
        raise ValueError("`hparams` must not be `None`.")

    layer_type = hparams["type"]
    if not is_str(layer_type) and not isinstance(layer_type, type):
        layer = layer_type
    else:
        layer_modules = ["torch.nn", "texar.torch.core", "texar.torch.custom"]
        layer_class: Type[nn.Module] = utils.check_or_get_class(
            layer_type, layer_modules)
        if isinstance(hparams, dict):
            if (layer_class.__name__ == "Linear" and
                    "in_features" not in hparams["kwargs"]):
                raise ValueError("\"in_features\" should be specified for "
                                 "\"torch.nn.{}\"".format(layer_class.__name__))
            elif (layer_class.__name__ in ["Conv1d", "Conv2d", "Conv3d"] and
                  "in_channels" not in hparams["kwargs"]):
                raise ValueError("\"in_channels\" should be specified for "
                                 "\"torch.nn.{}\"".format(layer_class.__name__))
            default_kwargs: Dict[str, Any] = {}
            default_hparams = {"type": layer_type, "kwargs": default_kwargs}
            hparams = HParams(hparams, default_hparams)

        # this case needs to be handled separately because nn.Sequential
        # does not accept kwargs
        if layer_type == "Sequential":
            names: List[str] = []
            layer = nn.Sequential()
            sub_hparams = hparams.kwargs.layers
            for hparam in sub_hparams:
                sub_layer = get_layer(hparam)
                name = utils.uniquify_str(sub_layer.__class__.__name__, names)
                names.append(name)
                layer.add_module(name=name, module=sub_layer)
        else:
            layer = utils.get_instance(layer_type, hparams.kwargs.todict(),
                                       layer_modules)

    if not isinstance(layer, nn.Module):
        raise ValueError("layer must be an instance of `torch.nn.Module`.")

    return layer


class MaxReducePool1d(nn.Module):
    r"""A subclass of :torch_nn:`Module`.
    Max Pool layer for 1D inputs. The same as :torch_nn:`MaxPool1d` except that
    the pooling dimension is entirely reduced (i.e., `pool_size=input_length`).
    """

    def forward(self,  # type: ignore
                input: torch.Tensor) -> torch.Tensor:
        output, _ = torch.max(input, dim=2)
        return output


class AvgReducePool1d(nn.Module):
    r"""A subclass of :torch_nn:`Module`.
    Avg Pool layer for 1D inputs. The same as :torch_nn:`AvgPool1d` except that
    the pooling dimension is entirely reduced (i.e., `pool_size=input_length`).
    """

    def forward(self,  # type: ignore
                input: torch.Tensor) -> torch.Tensor:
        return torch.mean(input, dim=2)


_POOLING_TO_REDUCE = {
    "MaxPool1d": "MaxReducePool1d",
    "AvgPool1d": "AvgReducePool1d",
    torch.nn.MaxPool1d: MaxReducePool1d,
    torch.nn.AvgPool1d: AvgReducePool1d
}


def get_pooling_layer_hparams(hparams: Union[HParams, Dict[str, Any]]) \
        -> Dict[str, Any]:
    r"""Creates pooling layer hyperparameters `dict` for :func:`get_layer`.

    If the :attr:`hparams` sets `'pool_size'` to `None`, the layer will be
    changed to the respective reduce-pooling layer. For example,
    :torch_docs:`torch.conv.MaxPool1d <nn.html#torch.nn.Conv1d>` is replaced
    with :class:`~texar.torch.core.MaxReducePool1d`.
    """
    if isinstance(hparams, HParams):
        hparams = hparams.todict()

    new_hparams = copy.copy(hparams)
    kwargs = new_hparams.get('kwargs', None)

    if kwargs and kwargs.get('kernel_size', None) is None:
        pool_type = hparams['type']
        new_hparams['type'] = _POOLING_TO_REDUCE.get(pool_type, pool_type)
        kwargs.pop('kernel_size', None)
        kwargs.pop('stride', None)
        kwargs.pop('padding', None)

    return new_hparams


class MergeLayer(nn.Module):
    r"""A subclass of :torch_nn:`Module`.
    A layer that consists of multiple layers in parallel. Input is fed to
    each of the parallel layers, and the outputs are merged with a
    specified mode.

    Args:
        layers (list, optional): A list of :torch_docs:`torch.nn.Module
            <nn.html#module>` instances, or a list of hyperparameter
            dictionaries each of which specifies `"type"` and `"kwargs"` of each
            layer (see the `hparams` argument of :func:`get_layer`).

            If `None`, this layer degenerates to a merging operator that merges
            inputs directly.
        mode (str): Mode of the merge op. This can be:

            - :attr:`'concat'`: Concatenates layer outputs along one dim.
              Tensors must have the same shape except for the dimension
              specified in `dim`, which can have different sizes.
            - :attr:`'elemwise_sum'`: Outputs element-wise sum.
            - :attr:`'elemwise_mul'`: Outputs element-wise product.
            - :attr:`'sum'`: Computes the sum of layer outputs along the
              dimension given by `dim`. For example, given `dim=1`,
              two tensors of shape `[a, b]` and `[a, c]` respectively
              will result in a merged tensor of shape `[a]`.
            - :attr:`'mean'`: Computes the mean of layer outputs along the
              dimension given in `dim`.
            - :attr:`'prod'`: Computes the product of layer outputs along the
              dimension given in `dim`.
            - :attr:`'max'`: Computes the maximum of layer outputs along the
              dimension given in `dim`.
            - :attr:`'min'`: Computes the minimum of layer outputs along the
              dimension given in `dim`.
            - :attr:`'and'`: Computes the `logical and` of layer outputs along
              the dimension given in `dim`.
            - :attr:`'or'`: Computes the `logical or` of layer outputs along
              the dimension given in `dim`.
            - :attr:`'logsumexp'`: Computes
              log(sum(exp(elements across the dimension of layer outputs)))
        dim (int): The dim to use in merging. Ignored in modes
            :attr:`'elemwise_sum'` and :attr:`'elemwise_mul'`.
    """

    _functions: Dict[str, Callable[[torch.Tensor, int], torch.Tensor]] = {
        "sum": torch.sum,
        "mean": torch.mean,
        "prod": torch.prod,
        "max": lambda tensors, dim: torch.max(tensors, dim)[0],
        "min": lambda tensors, dim: torch.min(tensors, dim)[0],
        "and": torch.all,
        "or": torch.any,
        "logsumexp": torch.logsumexp
    }

    def __init__(self, layers: Optional[List[nn.Module]] = None,
                 mode: str = 'concat', dim: Optional[int] = None):
        super().__init__()
        self._mode = mode
        self._dim = dim

        self._layers: Optional[nn.ModuleList] = None
        if layers is not None:
            if len(layers) == 0:
                raise ValueError(
                    "'layers' must be either None or a non-empty list.")
            self._layers = nn.ModuleList()
            for layer in layers:
                if isinstance(layer, nn.Module):
                    self._layers.append(layer)
                else:
                    self._layers.append(get_layer(hparams=layer))

    def forward(self, input: torch.Tensor) -> torch.Tensor:  # type: ignore
        r"""Feed input to every containing layer and merge the outputs.

        Args:
            input: The input tensor.

        Returns:
            The merged tensor.
        """
        layer_outputs: List[torch.Tensor]
        if self._layers is None:
            layer_outputs = input
            if not isinstance(layer_outputs, (list, tuple)):
                layer_outputs = [layer_outputs]
        else:
            layer_outputs = []
            for layer in self._layers:
                layer_output = layer(input)
                layer_outputs.append(layer_output)

        # the merge dimension cannot be determined until we get the output from
        # individual layers.
        # In case of reduce pooling operations, feature dim is removed and
        # channel dim is merged.
        # In non-reduce pooling operations, feature dim is merged.
        dim = self._dim if self._dim is not None else -1

        if self._mode == 'concat':
            outputs = torch.cat(tensors=layer_outputs, dim=dim)
        elif self._mode == 'elemwise_sum':
            outputs = layer_outputs[0]
            for i in range(1, len(layer_outputs)):
                outputs = torch.add(outputs, layer_outputs[i])
        elif self._mode == 'elemwise_mul':
            outputs = layer_outputs[0]
            for i in range(1, len(layer_outputs)):
                outputs = torch.mul(outputs, layer_outputs[i])
        elif self._mode in self._functions:
            _concat = torch.cat(tensors=layer_outputs, dim=dim)
            outputs = self._functions[self._mode](_concat, dim)
        else:
            raise ValueError("Unknown merge mode: '%s'" % self._mode)

        return outputs

    @property
    def layers(self) -> Optional[nn.ModuleList]:
        r"""The list of parallel layers.
        """
        return self._layers


class Flatten(nn.Module):
    r"""Flatten layer to flatten a tensor after convolution."""

    def forward(self,  # type: ignore
                input: torch.Tensor) -> torch.Tensor:
        return input.view(input.size()[0], -1)


class Identity(nn.Module):
    r"""Identity activation layer."""

    def forward(self,  # type: ignore
                input: torch.Tensor) -> torch.Tensor:
        return input



# Copyright 2019 The Texar Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Unit tests for Transformer encoder.
"""
import unittest

import torch

from texar.torch.modules.encoders import TransformerEncoder


class TransformerEncoderTest(unittest.TestCase):
    r"""Tests :class:`~texar.torch.modules.TransformerEncoder`
    """

    def setUp(self):
        self._batch_size = 2
        self._emb_dim = 512
        self._max_time = 7

    def test_trainable_variables(self):
        r"""Tests train_greedy
        """
        inputs = torch.rand(
            self._batch_size, self._max_time, self._emb_dim, dtype=torch.float)

        sequence_length = torch.randint(
            self._max_time, (self._batch_size,), dtype=torch.long)

        encoder = TransformerEncoder()

        outputs = encoder(inputs=inputs, sequence_length=sequence_length)

        # 6 blocks
        # -self multihead_attention: 4 dense without bias + 2 layer norm vars
        # -poswise_network: Dense with bias, Dense with bias + 2 layer norm vars
        # 2 output layer norm vars
        self.assertEqual(len(encoder.trainable_variables), 74)
        self.assertEqual(outputs.size(-1), encoder.output_size)

        hparams = {"use_bert_config": True}
        encoder = TransformerEncoder(hparams=hparams)

        # 6 blocks
        # -self multihead_attention: 4 dense without bias + 2 layer norm vars
        # -poswise_network: Dense with bias, Dense with bias + 2 layer norm vars
        # -output: 2 layer norm vars
        # 2 input layer norm vars
        outputs = encoder(inputs=inputs, sequence_length=sequence_length)
        self.assertEqual(len(encoder.trainable_variables), 74)
        self.assertEqual(outputs.size(-1), encoder.output_size)

    def test_encode(self):
        r"""Tests encoding.
        """
        inputs = torch.rand(
            self._batch_size, self._max_time, self._emb_dim, dtype=torch.float)

        sequence_length = torch.randint(
            self._max_time, (self._batch_size,), dtype=torch.long)

        encoder = TransformerEncoder()
        outputs = encoder(inputs=inputs, sequence_length=sequence_length)
        self.assertEqual(outputs.size(),
                         torch.Size((self._batch_size,
                                     self._max_time,
                                     self._emb_dim)))

        hparams = {"use_bert_config": True}
        encoder = TransformerEncoder(hparams=hparams)
        outputs = encoder(inputs=inputs, sequence_length=sequence_length)
        self.assertEqual(outputs.size(),
                         torch.Size((self._batch_size,
                                     self._max_time,
                                     self._emb_dim)))


if __name__ == "__main__":
    unittest.main()



# Copyright 2019 The Texar Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Unit tests for adv_losses.
"""

import unittest

import torch

from texar.torch.losses.adv_losses import binary_adversarial_losses


class AdvLossesTest(unittest.TestCase):
    """Tests adversarial losses.
    """

    def test_binary_adversarial_losses(self):
        """Tests :meth:`~texar.torch.losses.adv_losses.binary_adversarial_losse`.
        """
        batch_size = 16
        data_dim = 64
        real_data = torch.zeros(size=(batch_size, data_dim),
                                dtype=torch.float32)
        fake_data = torch.ones(size=(batch_size, data_dim),
                               dtype=torch.float32)
        const_logits = torch.zeros(size=(batch_size,), dtype=torch.float32)
        # Use a dumb discriminator that always outputs logits=0.
        gen_loss, disc_loss = binary_adversarial_losses(
            real_data, fake_data, lambda x: const_logits)
        gen_loss_2, disc_loss_2 = binary_adversarial_losses(
            real_data, fake_data, lambda x: const_logits, mode="min_fake")

        self.assertAlmostEqual(gen_loss.item(), -gen_loss_2.item())
        self.assertAlmostEqual(disc_loss.item(), disc_loss_2.item())


if __name__ == "__main__":
    unittest.main()



# Copyright 2019 The Texar Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Executor metrics for summaries.
"""

from collections import deque
from typing import Any, Deque, Optional, Sequence

import numpy as np
from torch.optim.optimizer import Optimizer

from texar.torch.run.metric.base_metric import StreamingMetric

__all__ = [
    "Average",
    "AveragePerplexity",
    "RunningAverage",
    "LR",
]


class Average(StreamingMetric[float, float]):
    r"""The average of a specific predicted value.

    Average is a :class:`~texar.torch.run.metric.StreamingMetric`, requires only
    predicted values. Average values are unbounded :class:`float` numbers. By
    default, lower values are better, but the behavior can be configured.

    Keyword Args:
        pred_name (str): Name of the predicted value. This will be used as the
            key to the dictionary returned by the model. Defaults to ``"loss"``.
        higher_is_better (bool, optional): If specified, the
            :attr:`higher_is_better` attribute for the instance is overwritten
            by the specified value. Defaults to `False`.
    """
    higher_is_better = False
    requires_label = False

    sum: float

    def __init__(self, *, pred_name: str = "loss",
                 higher_is_better: bool = False):
        super().__init__(pred_name=pred_name, higher_is_better=higher_is_better)

    def reset(self) -> None:
        super().reset()
        self.sum = 0.0

    def add(self, predicted: Sequence[float], _) -> None:
        self.count += len(predicted)
        self.sum += sum(predicted)

    def value(self) -> float:
        if self.count == 0:
            return 0.0
        return self.sum / self.count


class AveragePerplexity(Average):
    # TODO: Create a `WeightedAverage` class that takes `(value, weight)`
    #   and subclass that instead.
    higher_is_better = False

    def add(self, predicted: Sequence[float], _) -> None:
        super().add(np.exp(predicted), _)


class RunningAverage(StreamingMetric[float, float]):
    r"""The running average of a specific predicted value, i.e., the average
    computed over the most recent :attr:`queue_size` values.

    Running average is a :class:`~texar.torch.run.metric.StreamingMetric`,
    requires only predicted values. Running average values are unbounded
    :class:`float` numbers. By default, lower values are better, but the
    behavior can be configured.

    Keyword Args:
        queue_size (int): Size of the queue to keep history values. The running
            average is computed over the most recent :attr:`queue_size` values.
        pred_name (str): Name of the predicted value. This will be used as the
            key to the dictionary returned by the model. Defaults to ``"loss"``.
        higher_is_better (bool, optional): If specified, the
            :attr:`higher_is_better` attribute for the instance is overwritten
            by the specified value.
    """
    higher_is_better = False
    requires_label = False

    history: Deque[float]
    sum: float

    def __init__(self, queue_size: int, *, pred_name: str = "loss",
                 higher_is_better: bool = False):
        super().__init__(pred_name=pred_name, higher_is_better=higher_is_better)
        if not isinstance(queue_size, int) or queue_size <= 0:
            raise ValueError("'queue_size' must be a position integer")
        self.queue_size = queue_size

    def reset(self) -> None:
        super().reset()
        self.sum = 0.0
        self.history = deque()

    def add(self, predicted: Sequence[float], _) -> None:
        if len(predicted) >= self.queue_size:
            self.history = deque(predicted[-self.queue_size:])
            self.sum = sum(self.history)
        else:
            for _ in range(len(predicted) -
                           (self.queue_size - len(self.history))):
                self.sum -= self.history.popleft()
            self.sum += sum(predicted)
            self.history.extend(predicted)

    def value(self) -> float:
        if len(self.history) == 0:
            return 0.0
        return self.sum / len(self.history)


class LR(StreamingMetric[Any, float]):
    r"""The learning rate (LR) of the given optimizer. This is not exactly a
    metric, but rather a convenience object to print learning rates in log.

    LR is a :class:`~texar.torch.run.metric.StreamingMetric`, requires neither
    predicted values nor labels. LR values are unbounded :class:`float` numbers,
    with no clear definition of "better". Comparison of two learning rates are
    not meaningful.

    Keyword Args:
        optimizer: The optimizer instance.
        param_group (int, optional): Index of the parameter group to obtain the
            learning rate from. Defaults to 0. You don't need to specify this if
            the optimizer contains only one parameter group (e.g., constructed
            using :python:`optim_class(model.parameters())`.
    """

    requires_pred = False
    requires_label = False

    def __init__(self, optimizer: Optimizer, param_group: int = 0):
        super().__init__(pred_name=None)
        self.optimizer = optimizer
        self.group = param_group

    def add(self, _, __):
        pass

    def value(self) -> float:
        return self.optimizer.param_groups[self.group]['lr']  # type: ignore

    def better(self, cur: float, prev: float) -> Optional[bool]:
        # Always return `None` to indicate values are uncomparable.
        return None



# Copyright 2019 The Texar Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Utils for unit tests.
"""

import os
import unittest

__all__ = [
    "pretrained_test",
    "data_test",
    "external_library_test",
]


def define_skip_condition(flag: str, explanation: str):
    return unittest.skipUnless(
        os.environ.get(flag, 0) or os.environ.get('TEST_ALL', 0),
        explanation + f" Set `{flag}=1` or `TEST_ALL=1` to run.")


pretrained_test = define_skip_condition(
    'TEST_PRETRAINED', "Test requires loading pre-trained checkpoints.")
data_test = define_skip_condition(
    'TEST_DATA', "Test requires loading large data files.")


def external_library_test(name: str):
    import importlib
    try:
        importlib.import_module(name)
        return lambda x: x  # no changes
    except ImportError:
        return unittest.skip(
            f"Test requires external library {name} to be installed.")
